# This file is part of Xpra.
# Copyright (C) 2010 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# Platform-specific code for Win32 -- the parts that may import gtk.

import os
import sys
import types
from typing import Any
from collections.abc import Callable
from ctypes import (
    WinDLL, WinError, get_last_error,  # @UnresolvedImport
    CDLL, pythonapi, py_object,
    CFUNCTYPE, HRESULT, c_int, c_bool, c_char_p,
    create_string_buffer, POINTER, Structure, byref, addressof, sizeof,  # @UnresolvedImport
)
from ctypes.wintypes import HWND, DWORD, WPARAM, LPARAM, MSG, POINT, RECT, HGDIOBJ, LPCWSTR
from ctypes.util import find_library

from xpra.client.base import features
from xpra.exit_codes import ExitCode
from xpra.common import noop
from xpra.platform.win32 import constants as win32con, setup_console_event_listener
from xpra.platform.win32.window_hooks import Win32Hooks
from xpra.platform.win32.events import KNOWN_EVENTS, POWER_EVENTS
from xpra.platform.win32.common import (
    GetSystemMetrics, SetWindowLongA, GetWindowLongW,
    ClipCursor, GetCursorPos,
    GetDC, ReleaseDC,
    SendMessageA, GetMessageA, TranslateMessage, DispatchMessageA,
    FindWindowA,
    GetModuleHandleA,
    GetKeyState, GetWindowRect,
    GetDoubleClickTime,
    MonitorFromWindow, EnumDisplayMonitors,
    UnhookWindowsHookEx, CallNextHookEx, SetWindowsHookExA,
    GetDeviceCaps,
    GetIntSystemParametersInfo,
    GetUserObjectInformationA, OpenInputDesktop, CloseDesktop,
    GetMonitorInfo,
    GetKeyboardLayoutName,
)
from xpra.keyboard.common import KeyEvent
from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.util.env import envint, envbool
from xpra.util.io import get_util_logger
from xpra.log import Logger, consume_verbose_argv

log = Logger("win32")
grablog = Logger("win32", "grab")
screenlog = Logger("win32", "screen")
keylog = Logger("win32", "keyboard")
mouselog = Logger("win32", "mouse")

GLib = gi_import("GLib")

CONSOLE_EVENT_LISTENER = envbool("XPRA_CONSOLE_EVENT_LISTENER", True)
USE_NATIVE_TRAY = envbool("XPRA_USE_NATIVE_TRAY", True)
REINIT_VISIBLE_WINDOWS = envbool("XPRA_WIN32_REINIT_VISIBLE_WINDOWS", True)
SCREENSAVER_LISTENER_POLL_DELAY = envint("XPRA_SCREENSAVER_LISTENER_POLL_DELAY", 10)
APP_ID = os.environ.get("XPRA_WIN32_APP_ID", "Xpra")
MONITOR_DPI = envbool("XPRA_WIN32_MONITOR_DPI", True)

PyCapsule_GetPointer = pythonapi.PyCapsule_GetPointer
PyCapsule_GetPointer.restype = HGDIOBJ
PyCapsule_GetPointer.argtypes = [py_object]
log("PyCapsute_GetPointer=%s", PyCapsule_GetPointer)
GDK_DLL_NAME = "libgdk-3-0.dll"
gdk_dll = find_library(GDK_DLL_NAME)
if not gdk_dll:
    raise ImportError(f"ctypes cannot find {GDK_DLL_NAME!r}")
gdkdll = CDLL(gdk_dll)
gdk_win32_window_get_handle = gdkdll.gdk_win32_window_get_handle
gdk_win32_window_get_handle.argtypes = [HGDIOBJ]
gdk_win32_window_get_handle.restype = HWND
log("gdkdll=%s", gdkdll)

shell32 = WinDLL("shell32", use_last_error=True)
set_window_group: Callable = noop
try:
    from xpra.platform.win32 import propsys

    set_window_group = propsys.set_window_group
except ImportError as e:
    log("propsys missing", exc_info=True)
    log.warn("Warning: propsys support missing:")
    log.warn(" %s", e)
    log.warn(" window grouping is not available")

DwmGetWindowAttribute: Callable = noop
try:
    dwmapi = WinDLL("dwmapi", use_last_error=True)
    DwmGetWindowAttribute = dwmapi.DwmGetWindowAttribute
except Exception:
    # win XP:
    pass

WINDOW_HOOKS = envbool("XPRA_WIN32_WINDOW_HOOKS", True)
GROUP_LEADER = WINDOW_HOOKS and envbool("XPRA_WIN32_GROUP_LEADER", True)
UNDECORATED_STYLE = WINDOW_HOOKS and envbool("XPRA_WIN32_UNDECORATED_STYLE", True)
CLIP_CURSOR = WINDOW_HOOKS and envbool("XPRA_WIN32_CLIP_CURSOR", True)
# GTK3 is fixed, so we don't need this hook:
MAX_SIZE_HINT = WINDOW_HOOKS and envbool("XPRA_WIN32_MAX_SIZE_HINT", False)
LANGCHANGE = WINDOW_HOOKS and envbool("XPRA_WIN32_LANGCHANGE", True)
POLL_LAYOUT = envint("XPRA_WIN32_POLL_LAYOUT", 10)

FORWARD_WINDOWS_KEY = envbool("XPRA_FORWARD_WINDOWS_KEY", True)
WHEEL = envbool("XPRA_WHEEL", True)
WHEEL_DELTA = envint("XPRA_WIN32_WHEEL_DELTA", 120)
assert WHEEL_DELTA > 0

log("win32 gui settings: CONSOLE_EVENT_LISTENER=%s, USE_NATIVE_TRAY=%s, WINDOW_HOOKS=%s, GROUP_LEADER=%s",
    CONSOLE_EVENT_LISTENER, USE_NATIVE_TRAY, WINDOW_HOOKS, GROUP_LEADER)
log("win32 gui settings: UNDECORATED_STYLE=%s, CLIP_CURSOR=%s, MAX_SIZE_HINT=%s, LANGCHANGE=%s",
    UNDECORATED_STYLE, CLIP_CURSOR, MAX_SIZE_HINT, LANGCHANGE)
log("win32 gui settings: FORWARD_WINDOWS_KEY=%s, WHEEL=%s, WHEEL_DELTA=%s",
    FORWARD_WINDOWS_KEY, WHEEL, WHEEL_DELTA)


def do_init() -> None:
    from xpra.platform.win32.dpi import init_dpi
    init_dpi()
    if APP_ID:
        init_appid()
    try:
        from xpra.platform.win32.dwm_color import match_window_color
        match_window_color()
    except Exception:
        log.error("Error: failed to setup dwm window color matching", exc_info=True)


def init_appid() -> None:
    SetCurrentProcessExplicitAppUserModelID = shell32.SetCurrentProcessExplicitAppUserModelID
    SetCurrentProcessExplicitAppUserModelID.restype = HRESULT
    SetCurrentProcessExplicitAppUserModelID.argtypes = [LPCWSTR]
    if shell32.SetCurrentProcessExplicitAppUserModelID(APP_ID):
        log.warn("Warning: failed to set process app ID")


def use_stdin() -> bool:
    if os.environ.get("MSYSCON") or os.environ.get("CYGWIN"):
        return False
    stdin = sys.stdin
    if not stdin or not stdin.isatty():
        return False
    try:
        from xpra.platform.win32.common import GetStdHandle
        from xpra.platform.win32 import STD_INPUT_HANDLE, not_a_console, get_console_position
        hstdin = GetStdHandle(STD_INPUT_HANDLE)
        if not_a_console(hstdin):
            return False
        return get_console_position() != (-1, -1)
    except Exception:
        pass
    return True


def get_clipboard_native_class() -> str:
    # "xpra.clipboard.translated_clipboard.TranslatedClipboardProtocolHelper"
    return "xpra.platform.win32.clipboard.Win32Clipboard"


def get_native_notifier_classes() -> list[type]:
    try:
        from xpra.platform.win32.notifier import Win32_Notifier
        return [Win32_Notifier]
    except ImportError as e:
        log("no native notifier", exc_info=True)
        log.warn("Warning: cannot load native win32 notifier")
        log.warn(" %s", e)
    return []


def get_native_tray_classes() -> list[type]:
    c: list[type] = []
    if USE_NATIVE_TRAY:
        try:
            from xpra.platform.win32.tray import Win32Tray
            c.append(Win32Tray)
        except ImportError as e:
            log("no native tray", exc_info=True)
            log.warn("Warning: cannot load native win32 tray")
            log.warn(" %s", e)
    return c


def get_native_system_tray_classes(*_args):
    # Win32Tray cannot set the icon from data,
    # so it cannot be used for application trays
    return get_native_tray_classes()


def gl_check() -> str:
    # This is supposed to help `py2exe`
    # (must be done after we set up the `sys.path` in `platform.win32.paths`):
    try:
        from OpenGL.platform import win32
        assert win32
    except ImportError as e:
        get_util_logger().warn("gl_check()", exc_info=True)
        get_util_logger().warn("Warning: OpenGL bindings are missing")
        get_util_logger().warn(" %s", e)
        return "OpenGL bindings are missing"
    from xpra.platform.win32 import is_wine
    if is_wine():
        return "disabled when running under wine"
    return ""


def get_monitor_workarea_for_window(handle: int):
    try:
        monitor = MonitorFromWindow(handle, win32con.MONITOR_DEFAULTTONEAREST)
        mi = GetMonitorInfo(monitor)
        screenlog("get_monitor_workarea_for_window(%s) GetMonitorInfo(%s)=%s", handle, monitor, mi)
        # absolute workarea / monitor coordinates:
        # (all relative to 0,0 being top left)
        wx1, wy1, wx2, wy2 = mi['Work']
        mx1, my1, mx2, my2 = mi['Monitor']
        assert mx1 < mx2 and my1 < my2, "invalid monitor coordinates"
        # clamp to monitor, and make it all relative to monitor:
        rx1 = max(0, min(mx2 - mx1, wx1 - mx1))
        ry1 = max(0, min(my2 - my1, wy1 - my1))
        rx2 = max(0, min(mx2 - mx1, wx2 - mx1))
        ry2 = max(0, min(my2 - my1, wy2 - my1))
        assert rx1 < rx2 and ry1 < ry2, "invalid relative workarea coordinates"
        return rx1, ry1, rx2 - rx1, ry2 - ry1
    except Exception as e:
        log.warn("failed to query workareas: %s", e)
        return None


def get_window_handle(window) -> int:
    """ returns the win32 hwnd from a gtk.Window or gdk.Window """
    gdk_window = window
    try:
        gdk_window = window.get_window()
    except Exception:
        pass
    if not gdk_window:
        return 0
    gpointer = PyCapsule_GetPointer(gdk_window.__gpointer__, None)
    hwnd = gdk_win32_window_get_handle(gpointer)
    # log("get_window_handle(%s) gpointer=%#x, hwnd=%#x", gpointer, hwnd)
    return hwnd


def get_desktop_name() -> str:
    try:
        desktop = OpenInputDesktop(0, True, win32con.MAXIMUM_ALLOWED)
        if desktop:
            buf = create_string_buffer(128)
            r = GetUserObjectInformationA(desktop, win32con.UOI_NAME, buf, len(buf), None)
            if r != 0:
                desktop_name = buf.value.decode("latin1")
                return desktop_name
            CloseDesktop(desktop)
    except Exception as e:
        log.warn("Warning: failed to get desktop name")
        log.warn(" %s", e)
    return ""


def get_session_type() -> str:
    try:
        b = c_bool()
        retcode = dwmapi.DwmIsCompositionEnabled(byref(b))
        log("get_session_type() DwmIsCompositionEnabled()=%s (retcode=%s)", b.value, retcode)
        if retcode == 0 and b.value:
            return "aero"
    except (AttributeError, OSError):  # @UndefinedVariable
        # No windll, no dwmapi or no DwmIsCompositionEnabled function.
        log("get_session_type() failed to query DwmIsCompositionEnabled", exc_info=True)
    return ""


# alternative code:
#    try:
#        # Vista & 7 stuff
#        hwnd = GetDesktopWindow()
#        DwmGetWindowAttribute = dwmapi.DwmGetWindowAttribute
#        DWMWA_NCRENDERING_ENABLED = 1
#        b = BOOL()
#        DwmGetWindowAttribute(HWND(hwnd), DWORD(DWMWA_NCRENDERING_ENABLED), byref(b), sizeof(b))
#        #wx1,wy1,wx2,wy2 = rect.left, rect.top, rect.right, rect.bottom
#        log("DwmGetWindowAttribute: DWMWA_NCRENDERING_ENABLED(%i)=%s", hwnd, b)
#        if b:
#            return "aero"
#    except WindowsError as e:           #@UndefinedVariable
#        log("no DwmGetWindowAttribute: %s", e)
#    return ""


def win32_propsys_set_group_leader(self, leader):
    """ implements set group leader using propsys """
    hwnd = get_window_handle(self)
    if not hwnd:
        return
    try:
        log("win32_propsys_set_group_leader(%s)", leader)
        lhandle = get_window_handle(leader)
        assert lhandle
    except Exception:
        log("win32_propsys_set_group_leader(%s)", leader, exc_info=True)
        log.warn("Warning: no window handle for %s", leader)
        log.warn(" cannot set window grouping attribute")
        return
    if not lhandle:
        return
    try:
        log("win32 hooks: get_window_handle(%s)=%#x, set_group(%#x)", self, hwnd, lhandle)
        set_window_group(hwnd, lhandle)
    except Exception as e:
        log("set_window_group error", exc_info=True)
        log.error("Error: failed to set group leader")
        log.estr(e)


WS_NAMES: dict[int, str] = {
    win32con.WS_BORDER: "BORDER",
    win32con.WS_CAPTION: "CAPTION",
    win32con.WS_CHILD: "CHILD",
    win32con.WS_CHILDWINDOW: "CHILDWINDOW",
    win32con.WS_CLIPCHILDREN: "CLIPCHILDREN",
    win32con.WS_CLIPSIBLINGS: "CLIPSIBLINGS",
    win32con.WS_DISABLED: "DISABLED",
    win32con.WS_DLGFRAME: "DLGFRAME",
    win32con.WS_GROUP: "GROUP",
    win32con.WS_HSCROLL: "HSCROLL",
    win32con.WS_ICONIC: "ICONIC",
    win32con.WS_MAXIMIZE: "MAXIMIZE",
    win32con.WS_MAXIMIZEBOX: "MAXIMIZEBOX",
    win32con.WS_MINIMIZE: "MINIMIZE",
    win32con.WS_MINIMIZEBOX: "MINIMIZEBOX",
    win32con.WS_OVERLAPPED: "OVERLAPPED",
    win32con.WS_POPUP: "POPUP",
    win32con.WS_SIZEBOX: "SIZEBOX",
    win32con.WS_SYSMENU: "SYSMENU",
    win32con.WS_TABSTOP: "TABSTOP",
    win32con.WS_THICKFRAME: "THICKFRAME",
    win32con.WS_TILED: "TILED",
    win32con.WS_VISIBLE: "VISIBLE",
    win32con.WS_VSCROLL: "VSCROLL",
}


def style_str(style) -> str:
    return csv(s for c, s in WS_NAMES.items() if (c & style) == c)


def pointer_grab(window, *args) -> bool:
    hwnd = get_window_handle(window)
    grablog("pointer_grab%s window=%s, hwnd=%s", args, window, hwnd)
    if not hwnd:
        window._client.pointer_grabbed = None
        return False
    wrect = RECT()
    GetWindowRect(hwnd, byref(wrect))  # NOSONAR
    grablog("GetWindowRect(%i)=%s", hwnd, wrect)
    if DwmGetWindowAttribute != noop:
        # Vista & 7 stuff
        rect = RECT()
        DWMWA_EXTENDED_FRAME_BOUNDS = 9
        DwmGetWindowAttribute(HWND(hwnd), DWORD(DWMWA_EXTENDED_FRAME_BOUNDS), byref(rect), sizeof(rect))  # NOSONAR
        # wx1,wy1,wx2,wy2 = rect.left, rect.top, rect.right, rect.bottom
        grablog("DwmGetWindowAttribute: DWMWA_EXTENDED_FRAME_BOUNDS(%i)=%s", hwnd,
                (rect.left, rect.top, rect.right, rect.bottom))
    bx = GetSystemMetrics(win32con.SM_CXSIZEFRAME)
    by = GetSystemMetrics(win32con.SM_CYSIZEFRAME)
    top = by
    style = GetWindowLongW(hwnd, win32con.GWL_STYLE)
    if style & win32con.WS_CAPTION:
        top += GetSystemMetrics(win32con.SM_CYCAPTION)
    grablog(" window style=%s, SIZEFRAME=%s, top=%i", style_str(style), (bx, by), top)
    coords = wrect.left + bx, wrect.top + top, wrect.right - bx, wrect.bottom - by
    clip = RECT(*coords)
    r = ClipCursor(clip)
    grablog("ClipCursor%s=%s", coords, r)
    window._client.pointer_grabbed = window.wid
    return True


def pointer_ungrab(window, *args) -> bool:
    hwnd = get_window_handle(window)
    client = window._client
    grablog("pointer_ungrab%s window=%s, hwnd=%s, pointer_grabbed=%s",
            args, window, hwnd, client.pointer_grabbed)
    if not hwnd:
        return False
    grablog("ClipCursor(None)")
    ClipCursor(None)
    client.pointer_grabbed = None
    return True


def fixup_window_style(self, *_args) -> None:
    """ a fixup function we want to call from other places """
    hwnd = get_window_handle(self)
    if not hwnd:
        return
    try:
        # warning: accessing "_metadata" on the client window class is fugly..
        metadata = getattr(self, "_metadata", {})
        if metadata.get("modal", False):
            # window is not / no longer meant to be decorated
            # (this is what GTK does for modal windows - keep it consistent)
            return
        cur_style = GetWindowLongW(hwnd, win32con.GWL_STYLE)
        # re-add taskbar menu:
        style = cur_style
        if cur_style & win32con.WS_CAPTION:
            style |= win32con.WS_SYSMENU
            style |= win32con.WS_MAXIMIZEBOX
            style |= win32con.WS_MINIMIZEBOX
        # we can't tweak WS_MAXIMIZEBOX and WS_SIZEBOX
        # to hide the buttons
        # because GTK would then get confused
        # and paint the window contents at the wrong offset
        # hints = metadata.get("size-constraints")
        if style != cur_style:
            log("fixup_window_style() using %s (%#x) instead of %s (%#x) on window %#x with metadata=%s",
                style_str(style), style, style_str(cur_style), cur_style, hwnd, metadata)
            SetWindowLongA(hwnd, win32con.GWL_STYLE, style)
        else:
            log("fixup_window_style() unchanged style %s (%#x) on window %#x",
                style_str(style), style, hwnd)
        ws_visible = bool(style & win32con.WS_VISIBLE)
        client = self._client
        cur_ws_visible = getattr(self, "_ws_visible", True)
        iconified = getattr(self, "_iconified", False)
        been_mapped = getattr(self, "_been_mapped", False)
        log("fixup_window_style() ws_visible=%s (was %s), iconified=%s, been_mapped=%s", ws_visible, cur_ws_visible,
            iconified, been_mapped)
        if client and been_mapped and not iconified and ws_visible != cur_ws_visible:
            log("window changed visibility to: %s", ws_visible)
            setattr(self, "_ws_visible", ws_visible)
            if ws_visible:
                # with opengl, we need to re-create the window (PITA):
                if REINIT_VISIBLE_WINDOWS:
                    client.reinit_window(self.wid, self)
                self.send_control_refresh(False)
            else:
                self.send_control_refresh(True)
    except Exception:
        log.warn("failed to fixup window style", exc_info=True)


def set_decorated(self, decorated: bool) -> None:
    """ override method which ensures that we call
        fixup_window_style whenever decorations are toggled """
    self.__set_decorated(decorated)  # call the original saved method
    self.fixup_window_style()


def window_state_updated(window) -> None:
    """ fixup_window_style whenever the window state changes """
    log("window_state_updated(%s)", window)
    fixup_window_style(window)


def apply_maxsize_hints(window, hints: dict[str, Any]):
    """ extracts the max-size hints from the hints,
        and passes it to the win32hooks class which can implement it
        (as GTK does not honour it properly on win32)
    """
    workw, workh = 0, 0
    handle = get_window_handle(window)
    if not handle:
        return
    log("apply_maxsize_hints(%s, %s) handle=%#x", window, hints, handle)
    if not window.get_decorated():
        workarea = get_monitor_workarea_for_window(handle)
        log("using workarea as window size limit for undecorated window: %s", workarea)
        if workarea:
            workw, workh = workarea[2:4]
    thints = typedict(hints)
    minw = thints.intget("min_width", 0)
    minh = thints.intget("min_height", 0)
    maxw = thints.intget("max_width", 0)
    maxh = thints.intget("max_height", 0)
    if workw > 0 and workh > 0:
        # clamp to workspace for undecorated windows:
        if maxw > 0 and maxh > 0:
            maxw = min(workw, maxw)
            maxh = min(workh, maxh)
        else:
            maxw, maxh = workw, workh
    log("apply_maxsize_hints(%s, %s) found min: %sx%s, max: %sx%s", window, hints, minw, minh, maxw, maxh)
    if 0 < maxw < 32767 or 0 < maxh < 32767:
        window.win32hooks.max_size = (maxw or 32000), (maxh or 32000)
    elif window.win32hooks.max_size:
        # was set, clear it
        window.win32hooks.max_size = None
    if minw > 0 or minh > 0:
        window.win32hooks.min_size = minw, minh
    elif window.win32hooks.min_size:
        # was set, clear it:
        window.win32hooks.min_size = None
    if minw > 0 and minw == maxw and minh > 0 and minh == maxh:
        # fixed size, GTK can handle that
        return
    # remove them so GTK doesn't try to set attributes,
    # which would remove the maximize button:
    for x in ("min_width", "min_height", "max_width", "max_height"):
        hints.pop(x, None)
    window_state_updated(window)


def apply_geometry_hints(self, hints: dict):
    log("apply_geometry_hints(%s)", hints)
    apply_maxsize_hints(self, hints)
    return self.__apply_geometry_hints(hints)  # call the original saved method


def cache_pointer_offset(self, event):
    # this overrides the `window._get_pointer` method,
    # so we can cache the GTK position offset for synthetic wheel events
    gtk_x, gtk_y = event.x_root, event.y_root
    pos = POINT()
    GetCursorPos(addressof(pos))  # NOSONAR
    x, y = pos.x, pos.y
    self.win32_pointer_offset = gtk_x - x, gtk_y - y
    return gtk_x, gtk_y


def no_set_group(*_args):
    """ provide a dummy implementation """


def add_window_hooks(window) -> None:
    log("add_window_hooks(%s) WINDOW_HOOKS=%s, GROUP_LEADER=%s, UNDECORATED_STYLE=%s",
        window, WINDOW_HOOKS, GROUP_LEADER, UNDECORATED_STYLE)
    log(" MAX_SIZE_HINT=%s, MAX_SIZE_HINT=%s", MAX_SIZE_HINT, MAX_SIZE_HINT)
    if not WINDOW_HOOKS:
        # allows us to disable the win32 hooks for testing
        return
    try:
        gdk_window = window.get_window()
    except Exception:
        gdk_window = None
    if not gdk_window:
        # can't get a handle from a None value...
        return
    # at least provide a dummy method:
    gdk_window.set_group = no_set_group
    handle = get_window_handle(gdk_window)
    if not handle:
        log.warn("Warning: cannot add window hooks without a window handle!")
        return
    log("add_window_hooks(%s) gdk window=%s, hwnd=%#x", window, gdk_window, handle)

    if GROUP_LEADER:
        # MSWindows 7 onwards can use AppUserModel to emulate the group leader stuff:
        log("win32 hooks: set_window_group=%s", set_window_group)
        gdk_window.set_group = types.MethodType(win32_propsys_set_group_leader, gdk_window)
        log("hooked group leader override using %s", set_window_group)

    if UNDECORATED_STYLE:
        # OR windows never have any decorations or taskbar menu
        if not window._override_redirect:
            # the method to call to fix things up:
            window.fixup_window_style = types.MethodType(fixup_window_style, window)
            # override `set_decorated` so we can preserve the taskbar menu for undecorated windows
            window.__set_decorated = window.set_decorated
            window.set_decorated = types.MethodType(set_decorated, window)
            # override `after_window_state_updated` so we can re-add the missing style options
            # (somehow doing it from on_realize which calls add_window_hooks is not enough)
            window.connect("state-updated", window_state_updated)
            # call it at least once:
            window.fixup_window_style()

    if CLIP_CURSOR:
        window.pointer_grab = types.MethodType(pointer_grab, window)
        window.pointer_ungrab = types.MethodType(pointer_ungrab, window)

    if MAX_SIZE_HINT or LANGCHANGE or WHEEL:
        # glue code for gtk to win32 APIs:
        # add event hook class:
        win32hooks = Win32Hooks(handle)
        log("add_window_hooks(%s) added hooks for hwnd %#x: %s", window, handle, win32hooks)
        window.win32hooks = win32hooks
        win32hooks.setup()

        if MAX_SIZE_HINT:
            # save original geometry function:
            window.__apply_geometry_hints = window.apply_geometry_hints
            window.apply_geometry_hints = types.MethodType(apply_geometry_hints, window)
            # apply current max-size from hints, if any:
            if window.geometry_hints:
                apply_maxsize_hints(window, window.geometry_hints)

        if LANGCHANGE:
            def inputlangchange(_hwnd: int, _event: int, wParam: int, lParam: int) -> int:
                keylog("WM_INPUTLANGCHANGE: character set: %i, input locale identifier: %i", wParam, lParam)
                window.keyboard_layout_changed("WM_INPUTLANGCHANGE", wParam, lParam)
                return 0

            win32hooks.add_window_event_handler(win32con.WM_INPUTLANGCHANGE, inputlangchange)

        if WHEEL:
            VERTICAL = "vertical"
            HORIZONTAL = "horizontal"

            def handle_wheel(orientation: str, wParam: int, lParam: int):
                distance = wParam >> 16
                if distance > 2 ** 15:
                    # ie: 0xFF88 -> 0x78 (120)
                    distance = distance - 2 ** 16
                keys = wParam & 0xFFFF
                y = lParam >> 16
                x = lParam & 0xFFFF
                units = distance / WHEEL_DELTA
                client = getattr(window, "_client")
                wid = getattr(window, "_id", 0)
                mouselog(
                    "win32 mousewheel: orientation=%s, distance=%i, wheel-delta=%s, units=%.3f, new value=%.1f, keys=%#x, x=%i, y=%i, client=%s, wid=%i",
                    orientation, distance, WHEEL_DELTA, units, distance, keys, x, y, client, wid)
                if client and wid > 0:
                    if orientation == VERTICAL:
                        deltax = 0
                        deltay = units
                    else:
                        deltax = units
                        deltay = 0
                    pointer = window.get_mouse_position()
                    device_id = -1
                    client.wheel_event(device_id, wid, deltax, deltay, pointer)

            def mousewheel(_hwnd: int, _event: int, wParam: int, lParam: int) -> int:
                handle_wheel(VERTICAL, wParam, lParam)
                return 0

            def mousehwheel(_hwnd: int, _event: int, wParam: int, lParam: int) -> int:
                handle_wheel(HORIZONTAL, wParam, lParam)
                return 0

            win32hooks.add_window_event_handler(win32con.WM_MOUSEWHEEL, mousewheel)
            win32hooks.add_window_event_handler(win32con.WM_MOUSEHWHEEL, mousehwheel)


def remove_window_hooks(window) -> None:
    try:
        win32hooks = getattr(window, "win32hooks", None)
        if win32hooks:
            log("remove_window_hooks(%s) found %s", window, win32hooks)
            win32hooks.cleanup()
            window.win32hooks = None
    except Exception:
        log.error("remove_window_hooks(%s)", exc_info=True)


def get_xdpi() -> int:
    try:
        return _get_device_caps(win32con.LOGPIXELSX)
    except Exception as e:
        log.warn("failed to get xdpi: %s", e)
    return -1


def get_ydpi() -> int:
    try:
        return _get_device_caps(win32con.LOGPIXELSY)
    except Exception as e:
        log.warn("failed to get ydpi: %s", e)
    return -1


# those constants aren't found in win32con:
SPI_GETFONTSMOOTHING = 0x004A
SPI_GETFONTSMOOTHINGCONTRAST = 0x200C
SPI_GETFONTSMOOTHINGORIENTATION = 0x2012
FE_FONTSMOOTHINGORIENTATIONBGR = 0x0000
FE_FONTSMOOTHINGORIENTATIONRGB = 0x0001
FE_FONTSMOOTHINGORIENTATIONVBGR = 0x0002
FE_FONTSMOOTHINGORIENTATIONVRGB = 0x0003
SPI_GETFONTSMOOTHINGTYPE = 0x200A
FE_FONTSMOOTHINGCLEARTYPE = 0x0002
FE_FONTSMOOTHINGDOCKING = 0x8000
FE_ORIENTATION_STR: dict[int, str] = {
    FE_FONTSMOOTHINGORIENTATIONBGR: "BGR",
    FE_FONTSMOOTHINGORIENTATIONRGB: "RGB",
    FE_FONTSMOOTHINGORIENTATIONVBGR: "VBGR",
    FE_FONTSMOOTHINGORIENTATIONVRGB: "VRGB",
}
FE_FONTSMOOTHING_STR: dict[int, str] = {
    0: "Normal",
    FE_FONTSMOOTHINGCLEARTYPE: "ClearType",
}


def _add_SPI(info: dict[str, Any], constant: int, name: str, convert: Callable, default: Any = None) -> None:
    v = GetIntSystemParametersInfo(constant)
    if v is not None:
        info[name] = convert(v)
    elif default is not None:
        info[name] = default


def get_antialias_info() -> dict[str, Any]:
    info: dict[str, Any] = {}
    try:
        _add_SPI(info, SPI_GETFONTSMOOTHING, "enabled", bool)
        # "Valid contrast values are from 1000 to 2200. The default value is 1400."
        _add_SPI(info, SPI_GETFONTSMOOTHINGCONTRAST, "contrast", int)

        def orientation(v) -> str:
            return FE_ORIENTATION_STR.get(v, "unknown")

        _add_SPI(info, SPI_GETFONTSMOOTHINGORIENTATION, "orientation", orientation)

        def smoothing_type(v):
            return FE_FONTSMOOTHING_STR.get(v & FE_FONTSMOOTHINGCLEARTYPE, "unknown")

        _add_SPI(info, SPI_GETFONTSMOOTHINGTYPE, "type", smoothing_type)
        _add_SPI(info, SPI_GETFONTSMOOTHINGTYPE, "hinting", lambda v: bool(v & 0x2))
    except Exception as e:
        log.warn("failed to query antialias info: %s", e)
    return info


def get_mouse_config() -> dict[str, Any]:
    # not all are present in win32con?
    SM_CMOUSEBUTTONS = 43
    SM_CXDRAG = 68
    SM_CYDRAG = 69
    SM_MOUSEPRESENT = 19
    SM_MOUSEHORIZONTALWHEELPRESENT = 91
    SM_SWAPBUTTON = 23
    SM_MOUSEWHEELPRESENT = 75
    wheel_info = {
        "vertical": GetSystemMetrics(SM_MOUSEWHEELPRESENT),
        "horizontal": GetSystemMetrics(SM_MOUSEHORIZONTALWHEELPRESENT),
    }
    SPI_GETWHEELSCROLLLINES = 104
    SPI_GETWHEELSCROLLCHARS = 0x006C
    SPI_GETMOUSEVANISH = 4128
    # rate for each direction:
    _add_SPI(wheel_info, SPI_GETWHEELSCROLLLINES, "lines", int, 3)
    _add_SPI(wheel_info, SPI_GETWHEELSCROLLCHARS, "chars", int, 3)
    info: dict[str, Any] = {
        "present": bool(GetSystemMetrics(SM_MOUSEPRESENT)),
        "wheel": wheel_info,
        "buttons": GetSystemMetrics(SM_CMOUSEBUTTONS),
        "swap": bool(GetSystemMetrics(SM_SWAPBUTTON)),
        "drag": {
            "x": GetSystemMetrics(SM_CXDRAG),
            "y": GetSystemMetrics(SM_CYDRAG),
        },
    }
    _add_SPI(info, SPI_GETMOUSEVANISH, "vanish", bool, False)
    return info


def get_workarea() -> tuple[int, int, int, int] | None:
    # this is for x11 servers which can only use a single workarea,
    # calculate the total area:
    try:
        # first we need to find the absolute top-left and bottom-right corners
        # so we can make everything relative to 0,0
        monitors = []
        for m in EnumDisplayMonitors():
            mi = GetMonitorInfo(m)
            mx1, my1, mx2, my2 = mi['Monitor']
            monitors.append((mx1, my1, mx2, my2))
        minmx = min(x[0] for x in monitors)
        minmy = min(x[1] for x in monitors)
        maxmx = max(x[2] for x in monitors)
        maxmy = max(x[3] for x in monitors)
        screenlog("get_workarea() absolute total monitor area: %s", (minmx, minmy, maxmx, maxmy))
        screenlog(" total monitor dimensions: %s", (maxmx - minmx, maxmy - minmy))
        workareas = []
        for m in EnumDisplayMonitors():
            mi = GetMonitorInfo(m)
            # absolute workarea / monitor coordinates:
            wx1, wy1, wx2, wy2 = mi['Work']
            workareas.append((wx1, wy1, wx2, wy2))
        assert len(workareas) > 0
        minwx = min(w[0] for w in workareas)
        minwy = min(w[1] for w in workareas)
        maxwx = max(w[2] for w in workareas)
        maxwy = max(w[3] for w in workareas)
        # sanity checks:
        if minwx < minmx or minwy < minmy or maxwx > maxmx or maxwy > maxmy:
            raise ValueError("workspace %s is outside monitor space %s" % (
                (minwx, minwy, maxwx, maxwy), (minmx, minmy, maxmx, maxmy)
            ))
        # now make it relative to the monitor space:
        wx1 = minwx - minmx
        wy1 = minwy - minmy
        wx2 = maxwx - minmx
        wy2 = maxwy - minmy
        assert wx1 < wx2 and wy1 < wy2, "invalid workarea coordinates: %s" % ((wx1, wy1, wx2, wy2),)
        return wx1, wy1, wx2 - wx1, wy2 - wy1
    except Exception as e:
        screenlog("get_workarea()", exc_info=True)
        screenlog.warn("failed to query workareas: %s", e)
        return None


# ie: for a 60 pixel bottom bar on the second monitor at 1280x800:
# [(0,0,1920,1080), (0,0,1280,740)]
MONITORINFOF_PRIMARY = 1


def get_workareas() -> list[tuple[int, int, int, int]]:
    try:
        workareas: list[tuple[int, int, int, int]] = []
        for m in EnumDisplayMonitors():
            mi = GetMonitorInfo(m)
            screenlog("get_workareas() GetMonitorInfo(%s)=%s", m, mi)
            # absolute workarea / monitor coordinates:
            wx1, wy1, wx2, wy2 = mi['Work']
            mx1, my1, mx2, my2 = mi['Monitor']
            assert mx1 < mx2 and my1 < my2, "invalid monitor coordinates"
            # clamp to monitor, and make it all relative to monitor:
            rx1 = max(0, min(mx2 - mx1, wx1 - mx1))
            ry1 = max(0, min(my2 - my1, wy1 - my1))
            rx2 = max(0, min(mx2 - mx1, wx2 - mx1))
            ry2 = max(0, min(my2 - my1, wy2 - my1))
            assert rx1 < rx2 and ry1 < ry2, "invalid relative workarea coordinates"
            geom = rx1, ry1, rx2 - rx1, ry2 - ry1
            # GTK will return the PRIMARY monitor first,
            # so we have to do the same thing:
            if mi['Flags'] & MONITORINFOF_PRIMARY:
                workareas.insert(0, geom)
            else:
                workareas.append(geom)
        assert workareas
        screenlog("get_workareas()=%s", workareas)
        return workareas
    except Exception as e:
        screenlog.warn("failed to query workareas: %s", e)
        return []


def _get_device_caps(constant) -> int:
    dc = None
    try:
        dc = GetDC(None)
        return GetDeviceCaps(dc, constant)
    finally:
        if dc:
            ReleaseDC(None, dc)


def get_vrefresh() -> int:
    try:
        v = _get_device_caps(win32con.VREFRESH)
    except Exception as e:
        log("get_vrefresh()", exc_info=True)
        log.warn("Warning: failed to query the display vertical refresh rate:")
        log.warn(" %s", e)
        v = -1
    if v in (0, 1):
        # as per the docs:
        # "A vertical refresh rate value of 0 or 1 represents the display hardware's default refresh rate"
        return -1
    screenlog("get_vrefresh()=%s", v)
    return v


def get_double_click_time() -> int:
    try:
        return GetDoubleClickTime()
    except Exception as e:
        log.warn("failed to get double click time: %s", e)
        return 0


def get_double_click_distance() -> tuple[int, int]:
    try:
        return GetSystemMetrics(win32con.SM_CXDOUBLECLK), GetSystemMetrics(win32con.SM_CYDOUBLECLK)
    except Exception as e:
        log.warn("failed to get double click distance: %s", e)
        return -1, -1


def get_fixed_cursor_size() -> tuple[int, int]:
    try:
        w = GetSystemMetrics(win32con.SM_CXCURSOR)
        h = GetSystemMetrics(win32con.SM_CYCURSOR)
        return w, h
    except Exception as e:
        log.warn("failed to get window frame size information: %s", e)
        # best to try to use a limit anyway:
        return 32, 32


def get_cursor_size() -> int:
    w, h = get_fixed_cursor_size()
    return (w + h) // 2


def get_window_min_size() -> tuple[int, int]:
    return GetSystemMetrics(win32con.SM_CXMIN), GetSystemMetrics(win32con.SM_CYMIN)


# def get_window_max_size():
#    return 2**15-1, 2**15-1

def get_window_frame_sizes() -> dict[str, Any]:
    try:
        # normal resizable windows:
        rx = GetSystemMetrics(win32con.SM_CXSIZEFRAME)
        ry = GetSystemMetrics(win32con.SM_CYSIZEFRAME)
        # non-resizable windows:
        fx = GetSystemMetrics(win32con.SM_CXFIXEDFRAME)
        fy = GetSystemMetrics(win32con.SM_CYFIXEDFRAME)
        # min size:
        mx = GetSystemMetrics(win32con.SM_CXMIN)
        my = GetSystemMetrics(win32con.SM_CYMIN)
        # size of menu bar:
        m = GetSystemMetrics(win32con.SM_CYMENU)
        # border:
        b = GetSystemMetrics(win32con.SM_CYBORDER)
        # caption:
        c = GetSystemMetrics(win32con.SM_CYCAPTION)
        return {
            "normal": (rx, ry),
            "fixed": (fx, fy),
            "minimum": (mx, my),
            "menu-bar": m,
            "border": b,
            "caption": c,
            "offset": (rx, ry + c),
            # left, right, top, bottom:
            "frame": (rx, rx, ry + c, ry),
        }
    except Exception as e:
        log.warn("failed to get window frame size information: %s", e)
        return {}


def get_virtualscreenmetrics() -> tuple[int, int, int, int]:
    dx = GetSystemMetrics(win32con.SM_XVIRTUALSCREEN)
    dy = GetSystemMetrics(win32con.SM_YVIRTUALSCREEN)
    dw = GetSystemMetrics(win32con.SM_CXVIRTUALSCREEN)
    dh = GetSystemMetrics(win32con.SM_CYVIRTUALSCREEN)
    return dx, dy, dw, dh


def take_screenshot():
    # would be better to refactor the code..
    from xpra.platform.win32.gdi_screen_capture import GDICapture
    gdic = GDICapture()
    v = gdic.take_screenshot()
    gdic.clean()
    return v


def show_desktop(b) -> None:
    # not defined in win32con..
    MIN_ALL = 419
    MIN_ALL_UNDO = 416
    if bool(b):
        v = MIN_ALL
    else:
        v = MIN_ALL_UNDO
    try:
        root = FindWindowA("Shell_TrayWnd", None)
        assert root is not None, "cannot find 'Shell_TrayWnd'"
        SendMessageA(root, win32con.WM_COMMAND, v, 0)
    except Exception as e:
        log.warn("failed to call show_desktop(%s): %s", b, e)


def get_monitors_info(xscale=1, yscale=1) -> dict[int, Any]:
    from xpra.gtk.info import get_monitors_info
    monitors_info = get_monitors_info(xscale, yscale)
    if MONITOR_DPI:
        # try to get more precise data by querying the DPI using comtypes:
        from xpra.platform.win32.comtypes_util import CIMV2_Query
        with CIMV2_Query("SELECT * FROM Win32_DesktopMonitor") as res:
            index = 0
            for monitor in res:
                dminfo = {k: monitor.Properties_[k].Value for k in (
                    "DeviceID", "MonitorManufacturer", "MonitorType",
                    "ScreenWidth", "ScreenHeight",
                    "PixelsPerXLogicalInch", "PixelsPerYLogicalInch",
                )}
                log(f"Win32_DesktopMonitor {index}: {dminfo}")
                manufacturer = dminfo["MonitorManufacturer"]
                model = dminfo["MonitorType"]
                # find this monitor entry in the gtk info:
                mmatch = None
                for monitor_info in monitors_info.values():
                    if monitor_info.get("manufacturer") == manufacturer and monitor_info.get("model") == model:
                        mmatch = monitor_info
                        break
                if mmatch:
                    dpix = dminfo["PixelsPerXLogicalInch"]
                    dpiy = dminfo["PixelsPerYLogicalInch"]
                    # get the screen size from gtk because Win32_DesktopMonitor can be None!
                    width = dminfo["ScreenWidth"] or mmatch["geometry"][2]
                    height = dminfo["ScreenHeight"] or mmatch["geometry"][3]
                    if dpix > 0 and dpiy > 0 and width > 0 and height > 0:
                        mmatch.update({
                            "dpi-x": dpix,
                            "dpi-y": dpiy,
                            "width-mm": round(width * 25.4 / dpix),
                            "height-mm": round(height * 25.4 / dpiy),
                        })
                index += 1
    return monitors_info


TaskbarLib = None


def getTaskbar():
    # pylint: disable=import-outside-toplevel
    global TaskbarLib
    if TaskbarLib is None:
        taskbar_tlb = ""
        try:
            from xpra.platform.win32.comtypes_util import COMTYPES_ENABLED, QuietenLogging, find_tlb_file, comtypes_init
            if not COMTYPES_ENABLED:
                return None
            taskbar_tlb = find_tlb_file("TaskbarLib.tlb")
            if not taskbar_tlb:
                log.warn("Warning: 'TaskbarLib.tlb' was not found")
                log.warn(" taskbar integration cannot be enabled")
                TaskbarLib = False
                return None
            comtypes_init()
            with QuietenLogging():
                import comtypes.client as cc  # @UnresolvedImport
                cc.GetModule(taskbar_tlb)
                import comtypes.gen.TaskbarLib as tbl  # @UnresolvedImport
                TaskbarLib = tbl
                log(f"loaded {taskbar_tlb!r}: {TaskbarLib}")
        except Exception as e:
            log("getTaskbar()", exc_info=True)
            log.error(f"Error: failed to load taskbar library from {taskbar_tlb!r}")
            log.estr(e)
            TaskbarLib = False
    if not TaskbarLib:
        return None
    taskbar = cc.CreateObject("{56FDF344-FD6D-11d0-958A-006097C9A090}", interface=TaskbarLib.ITaskbarList3)
    taskbar.HrInit()
    return taskbar


def set_window_progress(window, pct: int) -> None:
    taskbar = getattr(window, "taskbar", None)
    if not taskbar:
        taskbar = getTaskbar()
        window.taskbar = taskbar
    if taskbar:
        handle = get_window_handle(window)
        taskbar.SetProgressValue(handle, max(0, min(100, pct)), 100)


WM_WTSSESSION_CHANGE = 0x02b1
WTS_CONSOLE_CONNECT = 0x1
WTS_CONSOLE_DISCONNECT = 0x2
WTS_REMOTE_CONNECT = 0x3
WTS_REMOTE_DISCONNECT = 0x4
WTS_SESSION_LOGON = 0x5
WTS_SESSION_LOGOFF = 0x6
WTS_SESSION_LOCK = 0x7
WTS_SESSION_UNLOCK = 0x8
WTS_SESSION_REMOTE_CONTROL = 0x9
WTS_SESSION_EVENTS: dict[int, str] = {
    WTS_CONSOLE_CONNECT: "CONSOLE CONNECT",
    WTS_CONSOLE_DISCONNECT: "CONSOLE_DISCONNECT",
    WTS_REMOTE_CONNECT: "REMOTE_CONNECT",
    WTS_REMOTE_DISCONNECT: "REMOTE_DISCONNECT",
    WTS_SESSION_LOGON: "SESSION_LOGON",
    WTS_SESSION_LOGOFF: "SESSION_LOGOFF",
    WTS_SESSION_LOCK: "SESSION_LOCK",
    WTS_SESSION_UNLOCK: "SESSION_UNLOCK",
    WTS_SESSION_REMOTE_CONTROL: "SESSION_REMOTE_CONTROL",
}


class ClientExtras:
    def __init__(self, client, _opts):
        self.client = client
        self._kh_warning = False
        self._console_handler_added = False
        self._screensaver_state = False
        self._screensaver_timer = 0
        self._exit = False
        if SCREENSAVER_LISTENER_POLL_DELAY > 0:
            def log_screensaver():
                v = bool(GetIntSystemParametersInfo(win32con.SPI_GETSCREENSAVERRUNNING))
                log("SPI_GETSCREENSAVERRUNNING=%s", v)
                if self._screensaver_state != v:
                    self._screensaver_state = v
                    if v:
                        self.client.suspend()
                    else:
                        self.client.resume()
                return True

            self._screensaver_timer = GLib.timeout_add(SCREENSAVER_LISTENER_POLL_DELAY * 1000, log_screensaver)
        if CONSOLE_EVENT_LISTENER:
            self._console_handler_added = setup_console_event_listener(self.handle_console_event, True)
        from xpra.platform.win32.events import get_win32_event_listener
        try:
            el = get_win32_event_listener(True)
            self._el = el
            if el:
                el.add_event_callback(win32con.WM_ACTIVATEAPP, self.activateapp)
                el.add_event_callback(win32con.WM_POWERBROADCAST, self.power_broadcast_event)
                el.add_event_callback(win32con.WM_MOVE, self.wm_move)
                el.add_event_callback(WM_WTSSESSION_CHANGE, self.session_change_event)
                el.add_event_callback(win32con.WM_INPUTLANGCHANGE, self.inputlangchange)
                el.add_event_callback(win32con.WM_WININICHANGE, self.inichange)
                el.add_event_callback(win32con.WM_ENDSESSION, self.end_session)
        except Exception as e:
            log.error("Error: cannot register focus and power callbacks:")
            log.estr(e)
        self.keyboard_hook_id = None
        self.keyboard_poll_timer: int = 0
        self.keyboard_id: int = self.get_keyboard_layout_id()
        if FORWARD_WINDOWS_KEY and features.windows:
            from xpra.util.thread import start_thread
            start_thread(self.init_keyboard_listener, "keyboard-listener", daemon=True)

    def ready(self) -> None:
        """ nothing specific to do here """

    def cleanup(self) -> None:
        log("ClientExtras.cleanup()")
        self._exit = True
        cha = self._console_handler_added
        if cha:
            self._console_handler_added = False
            # removing can cause crashes!?
            # setup_console_event_listener(self.handle_console_event, False)
        el = self._el
        if el:
            self._el = None
            el.cleanup()
        khid = self.keyboard_hook_id
        if khid:
            self.keyboard_hook_id = None
            UnhookWindowsHookEx(khid)
        kpt = self.keyboard_poll_timer
        if kpt:
            self.keyboard_poll_timer = 0
            GLib.source_remove(kpt)
        sst = self._screensaver_timer
        if sst:
            self._screensaver_timer = 0
            GLib.source_remove(sst)
        log("ClientExtras.cleanup() ended")
        # self.client = None

    def get_keyboard_layout_id(self) -> int:
        name_buf = create_string_buffer(win32con.KL_NAMELENGTH)
        if not GetKeyboardLayoutName(name_buf):
            return 0
        log(f"layout-name={name_buf.value!r}")
        try:
            # win32 API returns a hex string
            return int(name_buf.value, 16)
        except ValueError:
            log.warn("Warning: failed to parse keyboard layout code '%s'", name_buf.value)
        return 0

    def poll_layout(self) -> None:
        self.keyboard_poll_timer = 0
        klid = self.get_keyboard_layout_id()
        if klid and klid != self.keyboard_id:
            self.keyboard_id = klid
            self.client.window_keyboard_layout_changed()

    def init_keyboard_listener(self) -> None:
        class KBDLLHOOKSTRUCT(Structure):
            _fields_ = [
                ("vk_code", DWORD),
                ("scan_code", DWORD),
                ("flags", DWORD),
                ("time", c_int),
            ]

            def toString(self):
                return f"KBDLLHOOKSTRUCT({self.vk_code}, {self.scan_code}, {self.flags:x}, {self.time})"

        DOWN = [win32con.WM_KEYDOWN, win32con.WM_SYSKEYDOWN]
        # UP = [win32con.WM_KEYUP, win32con.WM_SYSKEYUP]
        ALL_KEY_EVENTS = {
            win32con.WM_KEYDOWN: "KEYDOWN",
            win32con.WM_SYSKEYDOWN: "SYSKEYDOWN",
            win32con.WM_KEYUP: "KEYUP",
            win32con.WM_SYSKEYUP: "SYSKEYUP",
        }

        def low_level_keyboard_handler(nCode: int, wParam: int, lParam: int):
            log("WH_KEYBOARD_LL: %s", (nCode, wParam, lParam))
            kh = getattr(self.client, "keyboard_helper", None)
            locked = getattr(kh, "locked", False)
            if POLL_LAYOUT and self.keyboard_poll_timer == 0 and not locked:
                self.keyboard_poll_timer = GLib.timeout_add(POLL_LAYOUT, self.poll_layout)
            if nCode < 0:
                # docs say we should not process this event:
                return CallNextHookEx(0, nCode, wParam, lParam)
            try:
                scan_code = lParam.contents.scan_code
                vk_code = lParam.contents.vk_code
                focused = self.client._focused
                # the keys we intercept before the OS:
                keyname = {
                    win32con.VK_LWIN: "Super_L",
                    win32con.VK_RWIN: "Super_R",
                    win32con.VK_TAB: "Tab",
                }.get(vk_code)
                modifiers: list[str] = []
                kh = self.client.keyboard_helper
                key_event_type = ALL_KEY_EVENTS.get(wParam)
                if self.client.keyboard_grabbed and focused and keyname and kh and kh.keyboard and key_event_type:
                    modifier_keycodes = kh.keyboard.modifier_keycodes
                    modifier_keys = kh.keyboard.modifier_keys
                    if keyname.startswith("Super"):
                        keycode = 0
                        # find the modifier keycode: (try the exact key we hit first)
                        for x in (keyname, "Super_L", "Super_R"):
                            keycodes = modifier_keycodes.get(x, [])
                            for k in keycodes:
                                # only interested in numeric keycodes:
                                try:
                                    keycode = int(k)
                                    break
                                except ValueError:
                                    pass
                            if keycode > 0:
                                break
                    else:
                        keycode = vk_code  # true for non-modifier keys only!
                    for vk, modkeynames in {
                        win32con.VK_NUMLOCK: ["Num_Lock"],
                        win32con.VK_CAPITAL: ["Caps_Lock"],
                        win32con.VK_CONTROL: ["Control_L", "Control_R"],
                        win32con.VK_SHIFT: ["Shift_L", "Shift_R"],
                    }.items():
                        if GetKeyState(vk):
                            for modkeyname in modkeynames:
                                mod = modifier_keys.get(modkeyname)
                                if mod:
                                    modifiers.append(mod)
                                    break
                    # keylog.info("keyboard helper=%s, modifier keycodes=%s", kh, modifier_keycodes)
                    grablog("vk_code=%s, scan_code=%s, event=%s, keyname=%s, keycode=%s, modifiers=%s, focused=%s",
                            vk_code, scan_code, ALL_KEY_EVENTS.get(wParam), keyname, keycode, modifiers, focused)
                    if keycode > 0:
                        key_event = KeyEvent()
                        key_event.keyname = keyname
                        key_event.pressed = wParam in DOWN
                        key_event.modifiers = modifiers
                        key_event.keyval = scan_code
                        key_event.keycode = keycode
                        key_event.string = ""
                        key_event.group = 0
                        grablog("detected '%s' key, sending %s", keyname, key_event)
                        self.client.keyboard_helper.send_key_action(focused, key_event)
                        # swallow this event:
                        return 1
            except Exception as e:
                keylog.error("Error: low level keyboard hook failed")
                keylog.estr(e)
            return CallNextHookEx(0, nCode, wParam, lParam)

        # Our low level handler signature.
        CMPFUNC = CFUNCTYPE(c_int, WPARAM, LPARAM, POINTER(KBDLLHOOKSTRUCT))
        # Convert the Python handler into C pointer.
        pointer = CMPFUNC(low_level_keyboard_handler)
        # Hook both key up and key down events for common keys (non-system).
        keyboard_hook_id = SetWindowsHookExA(win32con.WH_KEYBOARD_LL, pointer, GetModuleHandleA(None), 0)
        # Register to remove the hook when the interpreter exits:
        keylog("init_keyboard_listener() hook_id=%#x", keyboard_hook_id)
        msg = MSG()
        lpmsg = byref(msg)  # NOSONAR
        while not self._exit:
            ret = GetMessageA(lpmsg, 0, 0, 0)
            keylog("keyboard listener: GetMessage()=%s", ret)
            if ret == -1:
                raise WinError(get_last_error())
            if ret == 0:
                keylog("GetMessage()=0, exiting loop")
                return
            r = TranslateMessage(lpmsg)
            keylog("TranslateMessage(%#x)=%s", lpmsg, r)
            r = DispatchMessageA(lpmsg)
            keylog("DispatchMessageA(%#x)=%s", lpmsg, r)

    def wm_move(self, wParam: int, lParam: int) -> None:
        c = self.client
        log("WM_MOVE: %s/%s client=%s", wParam, lParam, c)
        if c:
            # this is not really a screen size change event,
            # but we do want to process it as such (see window reinit code)
            c.screen_size_changed()

    def end_session(self, wParam: int, lParam: int) -> None:
        log(f"WM_ENDSESSION({wParam}, {lParam})")
        c = self.client
        if not c:
            return
        ENDSESSION_CLOSEAPP = 0x1
        ENDSESSION_CRITICAL = 0x40000000
        ENDSESSION_LOGOFF = 0x80000000
        if (wParam & ENDSESSION_CLOSEAPP) and wParam:
            reason = "restart manager request"
        elif wParam & ENDSESSION_CRITICAL:
            reason = "application forced to shutdown"
        elif wParam & ENDSESSION_LOGOFF:
            reason = "logoff"
        else:
            return
        c.disconnect_and_quit(ExitCode.OK, reason)

    def session_change_event(self, event: int, session: int) -> None:
        event_name = WTS_SESSION_EVENTS.get(event) or str(event)
        log("WM_WTSSESSION_CHANGE: %s on session %#x", event_name, session)
        c = self.client
        if not c:
            return
        if event in (WTS_SESSION_LOGOFF, WTS_SESSION_LOCK):
            log("will freeze all the windows")
            c.freeze()
        elif event in (WTS_SESSION_LOGON, WTS_SESSION_UNLOCK):
            log("will unfreeze all the windows")
            # don't unfreeze directly from here,
            # as the system may not be fully usable yet (see #997)
            GLib.idle_add(c.unfreeze)

    def inputlangchange(self, wParam: int, lParam: int) -> None:
        keylog("WM_INPUTLANGCHANGE: %i, %i", wParam, lParam)
        self.poll_layout()

    def inichange(self, wParam: int, lParam: int) -> None:
        if lParam:
            log("WM_WININICHANGE: %#x=%s", lParam, c_char_p(lParam).value)
        else:
            log("WM_WININICHANGE: %i, %i", wParam, lParam)

    def activateapp(self, wParam: int, lParam: int) -> None:
        c = self.client
        log("WM_ACTIVATEAPP: %s/%s client=%s", wParam, lParam, c)
        if not c:
            return
        if not features.windows:
            return
        if wParam == 0:
            # our app has lost focus
            c.update_focus(0, False)
        # workaround for windows losing their style:
        for window in c._id_to_window.values():
            fixup_window_style = getattr(window, "fixup_window_style", None)
            if fixup_window_style:
                fixup_window_style()

    def power_broadcast_event(self, wParam: int, lParam: int) -> None:
        c = self.client
        log("WM_POWERBROADCAST: %s/%s client=%s", POWER_EVENTS.get(wParam, wParam), lParam, c)
        # maybe also "PBT_APMQUERYSUSPEND" and "PBT_APMQUERYSTANDBY"?
        if wParam == win32con.PBT_APMSUSPEND and c:
            c.suspend()
        # According to the documentation:
        # The system always sends a PBT_APMRESUMEAUTOMATIC message whenever the system resumes.
        elif wParam == win32con.PBT_APMRESUMEAUTOMATIC and c:
            c.resume()

    def handle_console_event(self, event: int) -> int:
        c = self.client
        event_name = KNOWN_EVENTS.get(event, event)
        log("handle_console_event(%s) client=%s, event_name=%s", event, c, event_name)
        info_events = [
            win32con.CTRL_C_EVENT,
            win32con.CTRL_LOGOFF_EVENT,
            win32con.CTRL_BREAK_EVENT,
            win32con.CTRL_SHUTDOWN_EVENT,
            win32con.CTRL_CLOSE_EVENT,
        ]
        if event in info_events:
            log.info("received console event %s", str(event_name).replace("_EVENT", ""))
        else:
            log.warn("unknown console event: %s", event_name)
        if c and event == win32con.CTRL_C_EVENT:
            log("calling=%s", c.signal_disconnect_and_quit)
            c.signal_disconnect_and_quit(0, "CTRL_C")
            return 1
        if c and event == win32con.CTRL_CLOSE_EVENT:
            c.signal_disconnect_and_quit(0, "CTRL_CLOSE")
            return 1
        return 0


def main():
    from xpra.platform import program_context
    with program_context("Platform-Events", "Platform Events Test"):
        consume_verbose_argv(sys.argv, "win32")
        log.info("Event loop is running")
        loop = GLib.MainLoop()

        from xpra.client.gui.fake_client import FakeClient
        fake_client = FakeClient()

        def signal_quit(*_args):
            loop.quit()

        fake_client.signal_disconnect_and_quit = signal_quit
        ClientExtras(fake_client, None)

        try:
            loop.run()
        except KeyboardInterrupt:
            log.info("exiting on keyboard interrupt")


if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.os_util import gi_import
from xpra.x11.desktop.base import DesktopServerBase
from xpra.x11.desktop.desktop_model import ScreenDesktopModel
from xpra.x11.bindings.randr import RandRBindings
from xpra.server import features
from xpra.gtk.error import xsync, xlog
from xpra.log import Logger

RandR = RandRBindings()

GLib = gi_import("GLib")
GObject = gi_import("GObject")

log = Logger("server")
geomlog = Logger("server", "window", "geometry")
screenlog = Logger("screen")


class XpraDesktopServer(DesktopServerBase):
    """
        A server class for RFB / VNC-like desktop displays,
        used with the "start-desktop" subcommand.
    """
    __gsignals__ = DesktopServerBase.__common_gsignals__

    def __init__(self):
        super().__init__()
        self.session_type = "desktop"
        self.gsettings_modified = {}
        self.root_prop_watcher = None
        self.resize_value = -1, -1

    def init_randr(self) -> None:
        super().init_randr()
        from xpra.x11.vfb_util import set_initial_resolution
        screenlog(f"init_randr() randr={self.randr}, initial-resolutions={self.initial_resolutions}")
        if not RandR.has_randr() or not self.initial_resolutions or not features.display:
            return
        res = self.initial_resolutions
        if len(res) > 1:
            log.warn(f"Warning: cannot set desktop resolution to {res}")
            log.warn(" multi monitor mode is not enabled")
            res = (res[0],)
            log.warn(f" using {res!r}")
        with xlog:
            set_initial_resolution(res, self.dpi or self.default_dpi)

    def configure_best_screen_size(self) -> tuple[int, int]:
        """ for the first client, honour desktop_mode_size if set """
        root_w, root_h = self.root_window.get_geometry()[2:4]
        if not self.randr:
            screenlog("configure_best_screen_size() no randr")
            return root_w, root_h
        sss = tuple(x for x in self._server_sources.values() if x.ui_client)
        if len(sss) != 1:
            screenlog.info(f"screen used by {len(sss)} clients:")
            return root_w, root_h
        ss = sss[0]
        requested_size = ss.desktop_mode_size
        if not requested_size:
            screenlog("configure_best_screen_size() client did not request a specific desktop mode size")
            return root_w, root_h
        w, h = requested_size
        screenlog("client requested desktop mode resolution is %sx%s (current server resolution is %sx%s)",
                  w, h, root_w, root_h)
        if w <= 0 or h <= 0 or w >= 32768 or h >= 32768:
            screenlog("configure_best_screen_size() client requested an invalid desktop mode size: %s", requested_size)
            return root_w, root_h
        return self.set_screen_size(w, h)

    def resize(self, w: int, h: int) -> None:
        geomlog("resize(%i, %i)", w, h)
        if not RandR.has_randr():
            geomlog.error("Error: cannot honour resize request,")
            geomlog.error(" no RandR support on this display")
            return
        # find the model:
        desktop_models = [window for window in self._id_to_window.values() if isinstance(window, ScreenDesktopModel)]
        if len(desktop_models) != 1:
            raise RuntimeError(f"found {desktop_models}, expected 1")
        geomlog(f"will resize {desktop_models}")
        desktop_models[0].resize(w, h)

    def get_server_mode(self) -> str:
        return "X11 desktop"

    def make_hello(self, source) -> dict[str, Any]:
        capabilities = super().make_hello(source)
        if "features" in source.wants:
            capabilities["desktop"] = True
        return capabilities

    def load_existing_windows(self) -> None:
        with xsync:
            model = ScreenDesktopModel(self.randr, self.randr_exact_size)
            model.setup()
            screenlog("adding root window model %s", model)
            super().do_add_new_window_common(1, model)
            model.managed_connect("client-contents-changed", self._contents_changed)
            model.managed_connect("resized", self.send_updated_screen_size)
            model.managed_connect("motion", self._motion_signaled)

    def send_updated_screen_size(self, model) -> None:
        # the vfb has been resized
        wid = self._window_to_id[model]
        x, y, w, h = model.get_geometry()
        geomlog("send_updated_screen_size(%s) geometry=%s", model, (x, y, w, h))
        for ss in self.window_sources():
            ss.resize_window(wid, model, w, h)
            ss.damage(wid, model, 0, 0, w, h)


GObject.type_register(XpraDesktopServer)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import time
import unittest
from contextlib import nullcontext

from xpra.os_util import POSIX
from xpra.util.objects import AdHocStruct
from unit.test_util import silence_info
from unit.server.mixins.servermixintest_util import ServerMixinTest


class ChildCommandMixinTest(ServerMixinTest):

    def test_command_server(self):
        try:
            from xpra.platform.posix.xdg_helper import log
            c = silence_info(log)
        except ImportError:
            c = nullcontext()
        with c:
            self.do_test_command_server()

    def do_test_command_server(self):
        from xpra.server.mixins import child_command
        opts = AdHocStruct()
        opts.exit_with_children = True
        opts.terminate_children = True
        opts.start_new_commands = True
        opts.start = []
        opts.start_child = []
        opts.start_late = []
        opts.start_child_late = []
        opts.start_after_connect = []
        opts.start_child_after_connect = []
        opts.start_on_connect = []
        opts.start_child_on_connect = []
        opts.start_on_last_client_exit = []
        opts.start_child_on_last_client_exit = []
        opts.exec_wrapper = None
        opts.start_env = []
        opts.source_start = []

        # pynotify can cause crashes,
        # probably due to threading issues?

        def noop():
            pass

        def _ChildCommandServer():
            ccs = child_command.ChildCommandServer()
            ccs.setup_menu_watcher = noop
            return ccs

        self._test_mixin_class(_ChildCommandServer, opts)
        if not POSIX:
            return
        # test creating a temp file:
        import tempfile
        tmpfile = os.path.join(tempfile.gettempdir(), "xpra-test-start-command-%s" % os.getpid())
        assert not os.path.exists(tmpfile)
        command = ("touch", tmpfile)
        with silence_info(child_command):
            self.handle_packet(("start-command", "test", command, True))
        time.sleep(1)
        info = self.mixin.get_info(self.protocol)
        commands = info.get("command")
        assert commands
        proc_info = commands.get(0)
        assert proc_info
        pid = proc_info.get("pid")
        assert pid
        assert os.path.exists(tmpfile)
        os.unlink(tmpfile)
        # test signals:
        with silence_info(child_command):
            self.handle_packet(("start-command", "sleep", "sleep 10", True))
        time.sleep(1)
        info = self.mixin.get_info(self.protocol)
        commands = info.get("command")
        assert commands
        proc_info = commands.get(1)
        assert proc_info
        pid = proc_info.get("pid")
        assert pid
        assert proc_info.get("name") == "sleep"
        assert proc_info.get("dead") is False
        # send it a SIGINT:
        with silence_info(child_command):
            self.handle_packet(("command-signal", pid, "SIGINT"))
        time.sleep(1)
        self.mixin.child_reaper.poll()
        info = self.mixin.get_info(self.protocol)
        commands = info.get("command")
        assert commands
        proc_info = commands.get(1)
        assert proc_info.get("dead") is True
        import signal
        assert proc_info.get("returncode") == -signal.SIGINT


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from queue import Empty
from collections.abc import Sequence
from weakref import WeakValueDictionary

from xpra.util.str_fn import Ellipsizer, print_nested_dict
from xpra.util.objects import typedict
from xpra.scripts.config import InitExit
from xpra.net.common import PacketType
from xpra.codecs.constants import VideoSpec, CodecStateException
from xpra.codecs.image import ImageWrapper, PlanarFormat
from xpra.codecs.remote.common import get_type, get_version, get_info, RemoteConnectionClient, RemoteCodec
from xpra.log import Logger

log = Logger("encoder", "remote")

ENCODINGS = tuple(x for x in os.environ.get("XPRA_REMOTE_ENCODINGS", "h264,vp8,vp9").split(",") if x)


def safe_dict(opts: dict | typedict) -> dict:
    # a simple dictionary suitable for sending via rencodeplus
    # (removes floats and other illegal values)
    return dict((k, v) for k, v in opts.items() if isinstance(k, str) and isinstance(v, (int, str, bool)))


class EncoderClient(RemoteConnectionClient):

    def __init__(self, options: dict):
        super().__init__(options)
        self.specs = {}
        self.encoders = WeakValueDictionary()
        self.encodings: Sequence[str] = ()

    def __repr__(self):
        return "EncoderClient(%s)" % self.uri

    def server_connection_cleanup(self):
        super().server_connection_cleanup()
        self.specs = {}
        self.encodings = ()

    def _process_encodings(self, packet: PacketType) -> None:
        log(f"{Ellipsizer(packet)!r}")
        specs = typedict(packet[1]).dictget("video") or {}
        self.specs = dict((k, v) for k, v in specs.items() if k in ENCODINGS)
        log("received specs=%s", Ellipsizer(specs))
        log("filtered specs:")
        print_nested_dict(self.specs, print_fn=log)
        encodings = tuple(self.specs.keys())
        self.encodings = tuple(set(encodings) & set(ENCODINGS))
        log("received encodings=%s", encodings)
        log("filtered encodings=%s", self.encodings)

    def request_close(self, seq: int, message="") -> None:
        if self.is_connected():
            self.send("context-close", seq, message)

    def request_context(self, encoder, encoding: str, width: int, height: int, src_format: str, options: dict) -> None:
        seq = encoder.generation
        codec_type = encoder.codec_type
        self.encoders[seq] = encoder
        sopts = safe_dict(options)
        self.send("context-request", seq, codec_type, encoding, width, height, src_format, sopts)

    def _process_context_response(self, packet: PacketType) -> None:
        seq = packet[1]
        ok = packet[2]
        message = "" if len(packet) < 4 else packet[3]
        info = {} if len(packet) < 5 else packet[4]
        encoder = self.encoders.get(seq)
        log(f"context-response: {seq}={encoder}, {ok=}, {message=!r}, {info=}")
        if not encoder:
            log(f"context response ignored, encoder {seq} not found!")
            return
        if ok:
            encoder.ready = True
        else:
            encoder.closed = True

    def compress(self, seq, image: ImageWrapper, options: typedict) -> None:
        log("compress%s", (seq, image, options))
        metadata = {}
        for attr in ("x", "y", "width", "height", "depth", "bytesperpixel", "planes"):
            metadata[attr] = int(getattr(image, f"get_{attr}")())
        for attr in ("rowstride", "pixel_format", "full_range"):
            metadata[attr] = getattr(image, f"get_{attr}")()
        pixels = image.get_pixels()
        nplanes = image.get_planes()
        send_opts = {}
        if self.lz4:
            send_opts["lz4"] = True
            from xpra.net.lz4.lz4 import compress
            if nplanes == PlanarFormat.PACKED:
                pixels = compress(pixels)
            else:
                pixels = [compress(plane) for plane in pixels]

        mmap_write_area = getattr(self, "mmap_write_area", None)
        if mmap_write_area:
            if nplanes == PlanarFormat.PACKED:
                mmap_data = mmap_write_area.write_data(pixels)
                log("sending image via mmap: %s", mmap_data)
            else:
                mmap_data = []
                for plane in range(nplanes):
                    plane_data = mmap_write_area.write_data(pixels[plane])
                    log("sending plane %i via mmap: %s", plane, plane_data)
                    mmap_data.append(plane_data)
            send_opts["chunks"] = tuple(mmap_data)
            pixels = b""
        self.send("context-compress", seq, metadata, pixels, safe_dict(options), send_opts)

    def _process_context_data(self, packet: PacketType) -> None:
        seq, bdata, client_options, reply_opts = packet[1:5]
        encoder = self.encoders.get(seq)
        if not encoder:
            log.error(f"Error: data unused, encoder {seq} not found!")
            return
        chunks = reply_opts.pop("chunks", ())
        if not bdata and chunks:
            mmap_read_area = getattr(self, "mmap_read_area", None)
            mmap_data, free = mmap_read_area.mmap_read(*chunks)
            bdata = bytes(mmap_data)
            free()
        log("server replied with %i bytes, client-options=%s", len(bdata), client_options)
        encoder.compressed_data(bdata, client_options)


server = None


def init_module(options: dict) -> None:
    log(f"encoder.init_module({options})")
    global server
    try:
        server = EncoderClient(options)
        server.connect()
        log("%s: %s, %s", get_type(), get_version(), get_info())
    except (InitExit, OSError, RuntimeError):
        log("failed to connect to server, no encodings available", exc_info=True)


def cleanup_module() -> None:
    log("remote.cleanup_module()")
    server.disconnect()
    server.cancel_schedule_connect()


def get_runtime_factor() -> float:
    return float(server.is_connected())


def make_spec(espec: dict) -> VideoSpec:
    codec_type = espec["codec_type"]

    class RemoteEncoder(Encoder):
        def __init__(self):
            super().__init__(server, codec_type)

    spec = VideoSpec(codec_class=RemoteEncoder, codec_type=f"remote-{codec_type}")
    for k, v in espec.items():
        if k == "codec_type":
            continue
        if not hasattr(spec, k):
            log.warn(f"Warning: unknown video spec attribute {k!r}")
            continue
        setattr(spec, k, v)
    spec.get_runtime_factor = get_runtime_factor
    return spec


def get_encodings() -> Sequence[str]:
    return tuple(server.specs.keys())


def get_specs() -> Sequence[VideoSpec]:
    # the `server.specs` are dictionaries,
    # which we need to convert to real `VideoSpec` objects:
    specs: Sequence[VideoSpec] = []
    for encoding, csc_specs in server.specs.items():
        for csc, especs in csc_specs.items():
            for espec in especs:
                log(f"remote: {encoding} + {csc}: {espec}")
                specs.append(make_spec(espec))
    log(f"remote.get_specs()={specs}")
    return tuple(specs)


class Encoder(RemoteCodec):

    def init_context(self, encoding: str, width: int, height: int, src_format: str, options: typedict) -> None:
        super().init_context(encoding, width, height, src_format, options)
        server.request_context(self, encoding, width, height, src_format, dict(options))

    def __repr__(self):
        if not self.pixel_format:
            return "remote.encoder(uninitialized)"
        return f"remote.encoder({self.pixel_format} - {self.width}x{self.height})"

    def get_src_format(self) -> str:
        return self.pixel_format

    def clean(self) -> None:
        super().clean()
        self.responses.put((b"", {}))
        server.request_close(self.generation, "encoder closed")

    def compress_image(self, image: ImageWrapper, options: typedict) -> tuple[bytes, dict]:
        if not server.is_connected():
            raise CodecStateException("not connected to encoder server")
        server.compress(self.generation, image, options)
        try:
            return self.responses.get(timeout=1)
        except Empty:
            log.warn("Warning: remote encoder timeout waiting for server response")
            log.warn(f" for {self.encoding!r} compression of {image}")
            self.closed = not server.is_connected()
            return b"", {}

    def compressed_data(self, bdata, client_options: dict) -> None:
        log(f"received {self.encoding!r} compressed data: {len(bdata)} bytes, {client_options=}")
        self.responses.put((bdata, client_options))



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Callable
from urllib.parse import unquote

from xpra.util.str_fn import Ellipsizer
from xpra.util.io import load_binary_file
from xpra.scripts.config import FALSE_OPTIONS
from xpra.net.common import HttpResponse
from xpra.platform.paths import get_icon_filename
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.log import Logger

log = Logger("http")


def invalid_path(uri: str) -> HttpResponse:
    log(f"invalid request path {uri!r}")
    return 404, {}, b""


def http_response(content, content_type: str = "text/plain") -> HttpResponse:
    if not content:
        return 404, {}, b""
    if isinstance(content, str):
        content = content.encode("latin1")
    return 200, {
        "Content-type": content_type,
        "Content-Length": len(content),
    }, content


def http_status_request() -> HttpResponse:
    return http_response("ready")


def http_icon_response(icon_type: str, icon_data: bytes) -> HttpResponse:
    log("http_icon_response%s", (icon_type, Ellipsizer(icon_data)))
    if not icon_data:
        icon_filename = get_icon_filename("noicon.png")
        icon_data = load_binary_file(icon_filename)
        icon_type = "png"
        log("using fallback transparent icon")
    if icon_type == "svg" and icon_data:
        from xpra.codecs.icon_util import svg_to_png  # pylint: disable=import-outside-toplevel
        # call svg_to_png via the main thread,
        # and wait for it to complete via an Event:
        icon: list[tuple[bytes, str]] = [(icon_data, icon_type)]
        from threading import Event
        event = Event()

        def convert() -> None:
            icon[0] = svg_to_png("", icon_data, 48, 48), "png"
            event.set()

        from xpra.os_util import gi_import
        GLib = gi_import("GLib")
        GLib.idle_add(convert)
        event.wait()
        icon_data, icon_type = icon[0]
    if icon_type in ("png", "jpeg", "svg", "webp"):
        mime_type = "image/" + icon_type
    else:
        mime_type = "application/octet-stream"
    return http_response(icon_data, mime_type)


def send_json_response(data) -> HttpResponse:
    import json  # pylint: disable=import-outside-toplevel
    return http_response(json.dumps(data), "application/json")


def _filter_display_dict(display_dict: dict[str, Any], *whitelist: str) -> dict[str, Any]:
    displays_info = {}
    for display, info in display_dict.items():
        displays_info[display] = {k: v for k, v in info.items() if k in whitelist}
    log("_filter_display_dict(%s)=%s", display_dict, displays_info)
    return displays_info


class HttpServer(StubServerMixin):
    """
    Mixin for servers that can handle http requests
    """
    PREFIX = "http"

    def __init__(self):
        self._http_scripts = {}

    def init(self, opts) -> None:
        http_scripts = opts.http_scripts
        if http_scripts.lower() in FALSE_OPTIONS:
            return
        script_options: dict[str, Callable[[str], HttpResponse]] = {
            "/Status": http_status_request,
            "/Info": self.http_info_request,
            "/Sessions": self.http_sessions_request,
            "/Displays": self.http_displays_request,
        }
        if self.menu_provider:
            # we have menu data we can expose:
            script_options |= {
                "/Menu": self.http_menu_request,
                "/MenuIcon": self.http_menu_icon_request,
                "/DesktopMenu": self.http_desktop_menu_request,
                "/DesktopMenuIcon": self.http_desktop_menu_icon_request,
            }
        if http_scripts.lower() in ("all", "*"):
            self._http_scripts = script_options
        else:
            for script in http_scripts.split(","):
                if not script.startswith("/"):
                    script = "/" + script
                handler = script_options.get(script)
                if not handler:
                    log.warn(f"Warning: unknown script {script!r}")
                else:
                    self._http_scripts[script] = handler
        log("init_http_scripts(%s)=%s", http_scripts, self._http_scripts)

    def cleanup(self) -> None:
        self._http_scripts = {}

    def get_info(self, _proto=None) -> dict[str, Any]:
        return {
            HttpServer.PREFIX: {"scripts": tuple(self._http_scripts.keys())},
        }

    def http_menu_request(self, _uri: str) -> HttpResponse:
        xdg_menu = self.menu_provider.get_menu_data(remove_icons=True)
        return send_json_response(xdg_menu or "not available")

    def http_desktop_menu_request(self, _uri: str) -> HttpResponse:
        xsessions = self.menu_provider.get_desktop_sessions(remove_icons=True)
        return send_json_response(xsessions or "not available")

    def http_menu_icon_request(self, uri: str) -> HttpResponse:
        parts = unquote(uri).split("/MenuIcon/", 1)
        # ie: "/menu-icon/a/b" -> ['', 'a/b']
        if len(parts) < 2:
            return invalid_path(uri)
        path = parts[1].split("/")
        # ie: "a/b" -> ['a', 'b']
        category_name = path[0]
        if len(path) < 2:
            # only the category is present
            app_name = ""
        else:
            app_name = path[1]
        log("http_menu_icon_request: category_name=%s, app_name=%s", category_name, app_name)
        icon_type, icon_data = self.menu_provider.get_menu_icon(category_name, app_name)
        return http_icon_response(icon_type, icon_data)

    def http_desktop_menu_icon_request(self, uri: str) -> HttpResponse:
        parts = unquote(uri).split("/DesktopMenuIcon/", 1)
        # ie: "/menu-icon/wmname" -> ['', 'sessionname']
        if len(parts) < 2:
            return invalid_path(uri)
        # in case the sessionname is followed by a slash:
        sessionname = parts[1].split("/")[0]
        log(f"http_desktop_menu_icon_request: {sessionname=}")
        icon_type, icon_data = self.menu_provider.get_desktop_menu_icon(sessionname)
        return http_icon_response(icon_type, icon_data)

    def http_displays_request(self, _uri: str) -> HttpResponse:
        displays = self.get_displays()
        displays_info = _filter_display_dict(displays, "state", "wmname", "xpra-server-mode")
        return send_json_response(displays_info)

    def get_displays(self) -> dict[str, Any]:
        from xpra.scripts.main import get_displays_info  # pylint: disable=import-outside-toplevel
        return get_displays_info(self.dotxpra)

    def http_sessions_request(self, _uri) -> HttpResponse:
        sessions = self.get_xpra_sessions()
        sessions_info = _filter_display_dict(sessions, "state", "username", "session-type", "session-name", "uuid")
        return send_json_response(sessions_info)

    def get_xpra_sessions(self) -> dict[str, Any]:
        from xpra.scripts.main import get_xpra_sessions  # pylint: disable=import-outside-toplevel
        return get_xpra_sessions(self.dotxpra)

    def http_info_request(self, _uri: str) -> HttpResponse:
        return send_json_response(self.get_http_info())

    def get_http_info(self) -> dict[str, Any]:
        return {
            "mode": self.get_server_mode(),
            "type": "Python",
            "uuid": self.uuid,
        }



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import logging
import unittest

from xpra.util.str_fn import csv
from xpra.common import noop
from xpra.codecs import loader
from xpra.codecs.constants import TransientCodecException

SUSPEND_CODEC_ERROR_LOGGING = os.environ.get("XPRA_SUSPEND_CODEC_ERROR_LOGGING", "1")=="1"
try:
    TEST_CODECS = os.environ.get("XPRA_TEST_CODECS").split(",")
except AttributeError:
    TEST_CODECS = loader.ALL_CODECS


class TestDecoders(unittest.TestCase):

    def test_all_codecs_found(self):
        # the self tests would swallow the exceptions and produce a warning:
        loader.RUN_SELF_TESTS = False
        # test them all:
        missing = []
        for codec_name in TEST_CODECS:
            loader.load_codec(codec_name)
            codec = loader.get_codec(codec_name)
            if not codec:
                missing.append(codec_name)
                continue
            try:
                # try to suspend error logging for full tests,
                # as those may cause errors
                log = getattr(codec, "log", None)
                if SUSPEND_CODEC_ERROR_LOGGING and log and not log.is_debug_enabled():
                    log.setLevel(logging.CRITICAL)
                init_module = getattr(codec, "init_module", noop)
                try:
                    init_module({})
                except Exception as e:
                    print("cannot initialize %s: %s" % (codec, e))
                    print(" test skipped")
                    continue

                selftest = getattr(codec, "selftest", noop)
                try:
                    selftest(True)
                except TransientCodecException as e:
                    print("ignoring TransientCodecException on %s : %s" % (codec, e))

                cleanup_module = getattr(codec, "cleanup_module", noop)
                cleanup_module()
            finally:
                if log:
                    log.setLevel(logging.DEBUG)
        if missing:
            print("Warning: the following codecs are missing and have not been tested:")
            print(f" {csv(missing)}")


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from importlib import import_module
from collections.abc import Sequence

from xpra.common import ALL_CLIPBOARDS
from xpra.client.base.stub_client_mixin import StubClientMixin
from xpra.platform.features import CLIPBOARD_WANT_TARGETS, CLIPBOARD_GREEDY, CLIPBOARD_PREFERRED_TARGETS, CLIPBOARDS
from xpra.platform.gui import get_clipboard_native_class
from xpra.net.common import PacketType, PacketElement
from xpra.net import compression
from xpra.scripts.config import FALSE_OPTIONS, TRUE_OPTIONS
from xpra.util.parsing import parse_simple_dict
from xpra.util.objects import typedict
from xpra.log import Logger

log = Logger("clipboard")

CLIPBOARD_CLASS = os.environ.get("XPRA_CLIPBOARD_CLASS", "")


def get_clipboard_helper_classes(clipboard_type: str) -> list[type]:
    ct = clipboard_type
    if ct and ct.lower() in FALSE_OPTIONS:
        return []
    # first add the platform specific one, (which may be None):
    clipboard_classes = [
        CLIPBOARD_CLASS,
        get_clipboard_native_class(),
    ]
    log("get_clipboard_helper_classes() unfiltered list=%s", clipboard_classes)
    if ct and ct.lower() != "auto" and ct.lower() not in TRUE_OPTIONS:
        # try to match the string specified:
        filtered = [x for x in clipboard_classes if x and x.lower().find(ct) >= 0]
        if not filtered:
            log.warn(f"Warning: no clipboard types matching {ct!r}")
            log.warn(" clipboard synchronization is disabled")
            return []
        log(" found %i clipboard types matching '%s'", len(filtered), ct)
        clipboard_classes = filtered
    # now try to load them:
    log("get_clipboard_helper_classes() options=%s", clipboard_classes)
    loadable = []
    for co in clipboard_classes:
        if not co:
            continue
        parts = co.split(".")
        mod = ".".join(parts[:-1])
        class_name = parts[-1]
        try:
            module = import_module(mod)
            helper_class = getattr(module, class_name)
            loadable.append(helper_class)
        except ImportError:
            log("cannot load %s", co, exc_info=True)
            log.warn(f"Warning: cannot load clipboard class {class_name!r} from {mod!r}")
            continue
    log("get_clipboard_helper_classes()=%s", loadable)
    return loadable


class ClipboardClient(StubClientMixin):
    """
    Utility mixin for clients that handle clipboard synchronization
    """
    __signals__ = ["clipboard-toggled"]
    PREFIX = "clipboard"

    def __init__(self):
        self.client_clipboard_type: str = ""
        self.client_clipboard_direction: str = "both"
        self.client_supports_clipboard: bool = False
        self.clipboard_enabled: bool = False
        self.server_clipboard_direction: str = "both"
        self.server_clipboard: bool = False
        self.server_clipboard_preferred_targets: Sequence[str] = ()
        self.server_clipboard_greedy: bool = False
        self.server_clipboard_want_targets: bool = False
        self.server_clipboard_selections: Sequence[str] = ()
        self.clipboard_helper = None
        self.local_clipboard_requests: int = 0
        self.remote_clipboard_requests: int = 0
        # only used with the translated clipboard class:
        self.local_clipboard: str = ""
        self.remote_clipboard: str = ""

    def init(self, opts) -> None:
        self.client_clipboard_type = opts.clipboard
        self.client_clipboard_direction = opts.clipboard_direction
        self.client_supports_clipboard = (opts.clipboard or "").lower() not in FALSE_OPTIONS
        self.remote_clipboard = opts.remote_clipboard
        self.local_clipboard = opts.local_clipboard

    def cleanup(self) -> None:
        ch = self.clipboard_helper
        log("ClipboardClient.cleanup() clipboard_helper=%s", ch)
        if ch:
            self.clipboard_helper = None
            with log.trap_error(f"Error on clipboard helper {ch} cleanup"):
                ch.cleanup()

    def get_info(self) -> dict[str, dict[str, Any]]:
        info: dict[str, Any] = {
            "client":
            {
                "enabled": self.clipboard_enabled,
                "type": self.client_clipboard_type,
                "direction": self.client_clipboard_direction,
            },
            "server":
            {
                "enabled": self.server_clipboard,
                "direction": self.server_clipboard_direction,
                "selections": self.server_clipboard_selections,
            },
            "requests":
            {
                "local": self.local_clipboard_requests,
                "remote": self.remote_clipboard_requests,
            },
        }
        return {"clipboard": info}

    def get_caps(self) -> dict[str, Any]:
        if not self.client_supports_clipboard:
            return {}
        caps: dict[str, Any] = {
            "": True,
            "enabled": True,
            "notifications": True,
            "selections": CLIPBOARDS,
            # buggy osx clipboards:
            "want_targets": CLIPBOARD_WANT_TARGETS,
            # buggy osx and win32 clipboards:
            "greedy": CLIPBOARD_GREEDY,
            "preferred-targets": CLIPBOARD_PREFERRED_TARGETS,
        }
        return {ClipboardClient.PREFIX: caps}

    def parse_server_capabilities(self, c: typedict) -> bool:
        try:
            from xpra import clipboard
            assert clipboard
        except ImportError:
            log.warn("Warning: clipboard module is missing")
            self.clipboard_enabled = False
            return True
        self.server_clipboard = c.boolget("clipboard")
        self.server_clipboard_direction = c.strget("clipboard-direction", "both")
        if self.server_clipboard_direction not in ("both", self.client_clipboard_direction):
            if self.client_clipboard_direction == "disabled":
                log("client clipboard is disabled")
            elif self.server_clipboard_direction == "disabled":
                log.warn("Warning: server clipboard synchronization is currently disabled")
                self.client_clipboard_direction = "disabled"
            elif self.client_clipboard_direction == "both":
                log.warn("Warning: server only supports '%s' clipboard transfers", self.server_clipboard_direction)
                self.client_clipboard_direction = self.server_clipboard_direction
            else:
                log.warn("Warning: incompatible clipboard direction settings")
                log.warn(" server setting: %s, client setting: %s",
                         self.server_clipboard_direction, self.client_clipboard_direction)
        self.server_clipboard_selections = c.strtupleget("clipboard.selections", ALL_CLIPBOARDS)
        log("server clipboard: supported=%s, selections=%s, direction=%s",
            self.server_clipboard, self.server_clipboard_selections, self.server_clipboard_direction)
        log("client clipboard: supported=%s, selections=%s, direction=%s",
            self.client_supports_clipboard, CLIPBOARDS, self.client_clipboard_direction)
        self.clipboard_enabled = self.client_supports_clipboard and self.server_clipboard
        self.server_clipboard_greedy = c.boolget("clipboard.greedy")
        self.server_clipboard_want_targets = c.boolget("clipboard.want_targets")
        self.server_clipboard_preferred_targets = c.strtupleget("clipboard.preferred-targets", ())
        log("server clipboard: greedy=%s, want_targets=%s, selections=%s",
            self.server_clipboard_greedy, self.server_clipboard_want_targets, self.server_clipboard_selections)
        log("parse_clipboard_caps() clipboard enabled=%s", self.clipboard_enabled)
        self.server_clipboard_preferred_targets = c.strtupleget("clipboard.preferred-targets", ())
        return True

    def process_ui_capabilities(self, caps: typedict) -> None:
        log("process_ui_capabilities() clipboard_enabled=%s", self.clipboard_enabled)
        if self.clipboard_enabled:
            ch = self.make_clipboard_helper()
            if not ch:
                log.warn("Warning: no clipboard support")
            self.clipboard_helper = ch
            self.clipboard_enabled = ch is not None
            log("clipboard helper=%s", ch)
            if self.clipboard_enabled:
                # tell the server about which selections we really want to sync with
                # (could have been translated, or limited if the client only has one, etc.)
                self.send_clipboard_selections(ch.get_remote_selections())
                ch.send_all_tokens()
        # ui may want to know this is now set:
        self.emit("clipboard-toggled")

    def init_authenticated_packet_handlers(self) -> None:
        self.add_legacy_alias("set-clipboard-enabled", f"{ClipboardClient.PREFIX}-status")
        for x in (
            "token", "request",
            "contents", "contents-none",
            "pending-requests", "enable-selections",
            "status",
        ):
            self.add_packet_handler(f"{ClipboardClient.PREFIX}-{x}", self._process_clipboard_packet, True)

    def make_clipboard_helper(self):
        """
            Try the various clipboard classes until we find one
            that loads ok. (some platforms have more options than others)
        """
        parts = self.client_clipboard_type.split(":", 1)
        clipboard_type = parts[0]
        options = {}
        if len(parts) > 1:
            options = parse_simple_dict(parts[1])
        clipboard_classes = get_clipboard_helper_classes(clipboard_type)
        log("make_clipboard_helper() options=%s", clipboard_classes)
        for helperclass in clipboard_classes:
            try:
                return self.setup_clipboard_helper(helperclass, options)
            except (ImportError, AttributeError) as e:
                log.error("Error: cannot instantiate %s:", helperclass)
                log.estr(e)
                del e
            except RuntimeError:
                log.error("Error: cannot instantiate %s", helperclass, exc_info=True)
        return None

    def _process_clipboard_packet(self, packet: PacketType) -> None:
        ch = self.clipboard_helper
        packet_type = packet[0]
        log("process_clipboard_packet: %s, helper=%s", packet_type, ch)
        if packet_type == "clipboard-status":
            self._process_clipboard_status(packet)
        elif ch:
            ch.process_clipboard_packet(packet)

    def _process_clipboard_status(self, packet: PacketType) -> None:
        clipboard_enabled, reason = packet[1:3]
        if self.clipboard_enabled != clipboard_enabled:
            log.info("clipboard toggled to %s by the server", ["off", "on"][int(clipboard_enabled)])
            log.info(" reason given: %r", reason)
            self.clipboard_enabled = bool(clipboard_enabled)
            self.emit("clipboard-toggled")

    def clipboard_toggled(self, *args) -> None:
        log("clipboard_toggled%s clipboard_enabled=%s, server_clipboard=%s",
            args, self.clipboard_enabled, self.server_clipboard)
        if self.server_clipboard:
            # non-legacy name would be "clipboard-status"
            self.send_now("set-clipboard-enabled", self.clipboard_enabled)
            if self.clipboard_enabled:
                ch = self.clipboard_helper
                assert ch is not None
                self.send_clipboard_selections(ch.get_remote_selections())
                ch.send_all_tokens()

    def send_clipboard_selections(self, selections: Sequence[str]) -> None:
        log("send_clipboard_selections(%s)", selections)
        self.send_now("clipboard-enable-selections", selections)

    def setup_clipboard_helper(self, helper_class, options: dict):
        log("setup_clipboard_helper(%s, %s)", helper_class, options)
        # first add the platform specific one, (which may be None):
        kwargs = options.copy()
        kwargs |= {
            # all the local clipboards supported:
            "clipboards.local": CLIPBOARDS,
            # all the remote clipboards supported:
            "clipboards.remote": self.server_clipboard_selections,
            "can-send": self.client_clipboard_direction in ("to-server", "both"),
            "can-receive": self.client_clipboard_direction in ("to-client", "both"),
            # the local clipboard we want to sync to (with the translated clipboard only):
            "clipboard.local": self.local_clipboard,
            # the remote clipboard we want to we sync to (with the translated clipboard only):
            "clipboard.remote": self.remote_clipboard,
        }
        log("setup_clipboard_helper() kwargs=%s", kwargs)

        hc = helper_class(self.clipboard_send, self.clipboard_progress, **kwargs)
        hc.set_preferred_targets(self.server_clipboard_preferred_targets)
        hc.set_greedy_client(self.server_clipboard_greedy)
        hc.set_want_targets_client(self.server_clipboard_want_targets)
        hc.enable_selections(self.server_clipboard_selections)
        return hc

    def clipboard_send(self, *parts: PacketElement) -> None:
        log("clipboard_send: %r", parts[0])
        if not self.clipboard_enabled:
            log("clipboard is disabled, not sending clipboard packet")
            return
        # replaces 'Compressible' items in a packet
        # with a subclass that calls self.compressed_wrapper
        # and which can therefore enable the brotli compressor:
        packet = list(parts)
        for i, v in enumerate(packet):
            if isinstance(v, compression.Compressible):
                packet[i] = self.compressible_item(v)
        self.send_now(*packet)

    def clipboard_progress(self, local_requests: int, remote_requests: int) -> None:
        log("clipboard_progress(%s, %s)", local_requests, remote_requests)
        if local_requests >= 0:
            self.local_clipboard_requests = local_requests
        if remote_requests >= 0:
            self.remote_clipboard_requests = remote_requests
        n = self.local_clipboard_requests+self.remote_clipboard_requests
        self.clipboard_notify(n)

    def compressible_item(self, compressible) -> compression.Compressible:
        """
            converts a 'Compressible' item into something that will
            call `self.compressed_wrapper` when compression is requested
            by the network encode thread.
        """
        client = self

        class ProtocolCompressible(compression.Compressible):
            __slots__ = ()

            def compress(self):
                return client.compressed_wrapper(self.datatype, self.data,
                                                 level=9, can_inline=False, brotli=True)
        return ProtocolCompressible(compressible.datatype, compressible.data)

    def clipboard_notify(self, n) -> None:
        log("clipboard_notify(%i)", n)



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import struct
from typing import Any
from collections.abc import Sequence

from xpra.platform.win32 import get_common_startmenu_dir, get_startmenu_dir
from xpra.log import Logger
from xpra.util.str_fn import print_nested_dict

log = Logger("exec")


def parse_link(content: bytes) -> str:
    # skip first 20 bytes (HeaderSize and LinkCLSID)
    # read the LinkFlags structure (4 bytes)
    lflags = struct.unpack('I', content[0x14:0x18])[0]
    position = 0x18
    # if the HasLinkTargetIDlist bit is set then skip the stored IDlist
    # structure and header
    if (lflags & 0x01) == 1:
        position = struct.unpack('H', content[0x4C:0x4E])[0] + 0x4E
    last_pos = position
    position += 0x04
    # get how long the file information is (LinkInfoSize)
    length = struct.unpack('I', content[last_pos:position])[0]
    # skip 12 bytes (LinkInfoHeaderSize, LinkInfoFlags, and VolumeIDOffset)
    position += 0x0C
    # go to the LocalBasePath position
    lbpos = struct.unpack('I', content[position:position + 0x04])[0]
    position = last_pos + lbpos
    # read the string at the given position of the determined length
    size = (length + last_pos) - position - 0x02
    temp = struct.unpack('c' * size, content[position:position + size])
    return ''.join([chr(ord(a)) for a in temp])


def read_link(path: str) -> str:
    try:
        with open(path, "rb") as stream:
            content = stream.read()
    except OSError:
        log(f"read_link({path!r})", exc_info=True)
        return ""
    try:
        return parse_link(content)
    except Exception as e:
        log("error parsing '%s': %s", path, e)
        return ""


def listdir(d) -> Sequence[str]:
    if not d:
        return ()
    try:
        return os.listdir(d)
    except PermissionError as e:
        log("listdir(%s)", d, exc_info=True)
        log.warn(f"Warning: cannot access directory {d!r}")
        log.warn(f" {e}")
        return ()


def load_subdir(d: str) -> dict[str, dict[str, str]]:
    # recurse down directories
    # and return a dictionary of entries
    menu = {}
    for x in listdir(d):
        if x.endswith(".ini"):
            continue
        name = os.path.join(d, x)
        if os.path.isdir(name):
            menu.update(load_subdir(name))
        elif name.endswith(".lnk"):
            exe = read_link(name)
            if exe:
                menu[x[:-4]] = {
                    "command": exe,
                }
    return menu


def load_dir(d: str) -> dict[str, dict[str, Any]]:
    log("load_dir(%s)", d)
    menu = {}
    for x in listdir(d):
        log(f" {x}")
        if x.endswith(".ini"):
            continue
        name = os.path.join(d, x)
        if os.path.isdir(name):
            subdirmenu = load_subdir(name)
            if subdirmenu:
                menu[x] = {
                    "Entries": subdirmenu,
                }
        elif name.endswith(".lnk"):
            # add them to the "Shortcuts" submenu:
            exe = read_link(name)
            if exe:
                menu.setdefault("Shortcuts", {}).setdefault("Entries", {})[x[:-4]] = {
                    "command": exe,
                }
    return menu


def load_menu() -> dict[str, dict[str, Any]]:
    menu = {}
    for menu_dir in (get_common_startmenu_dir(), get_startmenu_dir()):
        # ie: "C:\ProgramData\Microsoft\Windows\Start Menu"
        for x in listdir(menu_dir):
            subdir = os.path.join(menu_dir, x)
            if os.path.isdir(subdir):
                # ie: "C:\ProgramData\Microsoft\Windows\Start Menu\Programs"
                m = load_dir(subdir)
                if not m:
                    continue
                # TODO: recursive merge
                for k, v in m.items():
                    ev = menu.get(k)
                    if isinstance(ev, dict):
                        ev.update(v)
                    else:
                        menu[k] = v
    return menu


def main():
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    with program_context("menu-helper", "Menu Helper"):
        menu = load_menu()
        print_nested_dict(menu)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import signal
from time import monotonic

from xpra.os_util import gi_import
from xpra.exit_codes import ExitValue
from xpra.gtk.signals import register_os_signals
from xpra.util.objects import typedict, AdHocStruct
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import scaled_image, label
from xpra.gtk.pixbuf import get_icon_pixbuf, get_pixbuf_from_data
from xpra.log import Logger, enable_debug_for

log = Logger("util")

GLib = gi_import("GLib")
Gtk = gi_import("Gtk")

_instance = None


def getServerCommandsWindow(client):
    global _instance
    if _instance is None:
        _instance = ServerCommandsWindow(client)
    return _instance


class ServerCommandsWindow:

    def __init__(self, client):
        assert client
        self.client = client
        self.populate_timer = 0
        self.commands_info = {}
        self.contents = None
        self.window = Gtk.Window()
        self.window.set_border_width(20)
        self.window.connect("delete-event", self.close)
        self.window.set_default_size(400, 150)
        self.window.set_title("Server Commands")

        icon_pixbuf = get_icon_pixbuf("list.png")
        if icon_pixbuf:
            self.window.set_icon(icon_pixbuf)
        self.window.set_position(Gtk.WindowPosition.CENTER)

        vbox = Gtk.VBox(homogeneous=False, spacing=0)
        vbox.set_spacing(10)

        self.alignment = Gtk.Alignment(xalign=0.5, yalign=0.5, xscale=1.0, yscale=1.0)
        vbox.pack_start(self.alignment, expand=True, fill=True)

        # Buttons:
        hbox = Gtk.HBox(homogeneous=False, spacing=20)
        vbox.pack_start(hbox)

        def btn(label, tooltip, callback, icon_name=None):
            b = self.btn(label, tooltip, callback, icon_name)
            hbox.pack_start(b)

        if self.client.server_start_new_commands:
            btn("Start New", "Run a command on the server", self.client.show_start_new_command, "forward.png")
        btn("Close", "", self.close, "quit.png")

        add_close_accel(self.window, self.close)
        vbox.show_all()
        self.window.vbox = vbox
        self.window.add(vbox)

    def btn(self, label, tooltip, callback, icon_name=None):
        btn = Gtk.Button(label=label)
        settings = btn.get_settings()
        settings.set_property('gtk-button-images', True)
        btn.set_tooltip_text(tooltip)
        btn.connect("clicked", callback)
        icon = get_icon_pixbuf(icon_name)
        if icon:
            btn.set_image(scaled_image(icon, 24))
        return btn

    def populate_table(self):
        commands_info = typedict(self.client.server_last_info).dictget("commands", {})
        if self.commands_info != commands_info and commands_info:
            log("populate_table() new commands_info=%s", commands_info)
            self.commands_info = commands_info
            if self.contents:
                self.alignment.remove(self.contents)
            grid = Gtk.Grid()

            def l(s=""):  # noqa: E743
                widget = label(s)
                widget.set_margin_start(5)
                widget.set_margin_end(5)
                return widget

            headers = ["", "PID", "Command", "Exit Code"]
            if self.client.server_commands_signals:
                headers.append("Send Signal")
            for i, text in enumerate(headers):
                grid.attach(l(text), i, 0, 1, 1)
            for row, procinfo in enumerate(self.commands_info.values()):
                if not isinstance(procinfo, dict):
                    continue
                # some records aren't procinfos:
                pi = typedict(procinfo)
                command = pi.strtupleget("command")
                pid = pi.intget("pid", 0)
                returncode: int | None = pi.intget("returncode") if "returncode" in pi else None
                if pid > 0 and command:
                    cmd_str = " ".join(command)
                    rstr = ""
                    if returncode is not None:
                        rstr = "%s" % returncode
                    # find the windows matching this pid
                    windows = ()
                    from xpra.client.base import features
                    if features.windows:
                        windows = tuple(w for w in self.client._id_to_window.values()
                                        if getattr(w, "_metadata", {}).get("pid") == pid)
                        log(f"windows matching pid={pid}: {windows}")
                    icon = label()
                    if windows:
                        try:
                            icons = tuple(getattr(w, "_current_icon", None) for w in windows)
                            icons = tuple(x for x in icons if x is not None)
                            log("icons: %s", icons)
                            if icons:
                                from PIL import Image  # @UnresolvedImport pylint: disable=import-outside-toplevel
                                try:
                                    LANCZOS = Image.Resampling.LANCZOS
                                except AttributeError:
                                    LANCZOS = Image.LANCZOS
                                img = icons[0].resize((24, 24), LANCZOS)
                                has_alpha = img.mode == "RGBA"
                                width, height = img.size
                                rowstride = width * (3 + int(has_alpha))
                                pixbuf = get_pixbuf_from_data(img.tobytes(), has_alpha, width, height, rowstride)
                                icon = Gtk.Image()
                                icon.set_from_pixbuf(pixbuf)
                        except Exception:
                            log("failed to get window icon", exc_info=True)
                    widgets = [icon, l(f"{pid}"), l(cmd_str), l(rstr)]
                    if self.client.server_commands_signals:
                        if returncode is None:
                            widgets.append(self.signal_button(pid))
                    for i, widget in enumerate(widgets):
                        grid.attach(widget, i, 1 + row, 1, 1)
            self.alignment.add(grid)
            grid.show_all()
            self.contents = grid
        self.client.send_info_request()
        return True

    def signal_button(self, pid):
        hbox = Gtk.HBox()
        combo = Gtk.ComboBoxText()
        for x in self.client.server_commands_signals:
            combo.append_text(x)

        def send(*_args):
            a = combo.get_active()
            if a >= 0:
                signame = self.client.server_commands_signals[a]
                self.client.send("command-signal", pid, signame)

        b = self.btn("Send", None, send, "forward.png")
        hbox.pack_start(combo)
        hbox.pack_start(b)
        return hbox

    def schedule_timer(self):
        if not self.populate_timer:
            self.populate_table()
            self.populate_timer = GLib.timeout_add(1000, self.populate_table)

    def cancel_timer(self):
        pt = self.populate_timer
        if pt:
            self.populate_timer = 0
            GLib.source_remove(pt)

    def show(self):
        log("show()")
        self.window.show_all()
        self.window.present()
        self.schedule_timer()

    def close(self, *args):
        log("close%s", args)
        if self.window:
            self.window.hide()
        self.cancel_timer()
        return True

    def destroy(self, *args):
        log("close%s", args)
        self.cancel_timer()
        if self.window:
            self.window.close()
            self.window = None

    def run(self) -> ExitValue:
        log("run()")
        Gtk.main()
        log("run() Gtk.main done")
        return 0

    def quit(self, *args):
        log("quit%s", args)
        self.close()
        Gtk.main_quit()


def main():  # pragma: no cover
    from xpra.platform import program_context
    from xpra.platform.gui import ready as gui_ready, init as gui_init
    gui_init()
    with program_context("Start-New-Command", "Start New Command"):
        if "-v" in sys.argv:
            enable_debug_for("util")

        client = AdHocStruct()
        client.server_last_info_time = monotonic()
        commands_info = {
            0: {
                'returncode': None, 'name': 'xterm', 'pid': 542, 'dead': False,
                'ignore': True, 'command': ('xterm',), 'forget': False,
            },
            'start-child': (),
            'start-new': True,
            'start-after-connect-done': True,
            'start': ('xterm',),
            'start-after-connect': (),
            'start-child-on-connect': (),
            'exit-with-children': False,
            'start-child-after-connect': (),
            'start-on-connect': (),
        }
        client.server_last_info = {"commands": commands_info}
        client.server_start_new_commands = True
        client.server_commands_signals = ("SIGINT", "SIGTERM", "SIGUSR1")

        def nosend(*_args):
            """
            this is for testing only - we are not connected to a server
            """

        client.send_info_request = nosend
        client.send = nosend
        window1 = AdHocStruct()
        window1._metadata = {"pid": 542}  # pylint: disable=protected-access
        client._id_to_window = {  # pylint: disable=protected-access
            1: window1
        }

        def show_start_new_command(*_args):
            from xpra.gtk.dialogs.start_new_command import getStartNewCommand
            getStartNewCommand(None).show()

        client.show_start_new_command = show_start_new_command

        app = ServerCommandsWindow(client)
        app.close = app.quit
        register_os_signals(app.quit)
        try:
            gui_ready()
            app.show()
            return app.run()
        except KeyboardInterrupt:
            return 128 + int(signal.SIGINT)


if __name__ == "__main__":  # pragma: no cover
    v = main()
    sys.exit(v)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
from typing import Any
from collections.abc import Sequence
from importlib import import_module

from xpra.platform.features import (
    CLIPBOARDS, CLIPBOARD_PREFERRED_TARGETS,
    CLIPBOARD_WANT_TARGETS, CLIPBOARD_GREEDY,
)
from xpra.os_util import gi_import
from xpra.util.str_fn import csv
from xpra.net.common import PacketType, PacketElement
from xpra.scripts.config import FALSE_OPTIONS
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("clipboard")


class ClipboardServer(StubServerMixin):
    """
    Mixin for servers that handle clipboard synchronization.
    """
    PREFIX = "clipboard"

    def __init__(self):
        self.clipboard = False
        self.clipboard_direction = "none"
        self.clipboard_filter_file = None
        self._clipboard_helper = None
        self._clipboard_client = None
        self._clipboards: Sequence[str] = ()

    def init(self, opts) -> None:
        self.clipboard = (opts.clipboard or "").lower() not in FALSE_OPTIONS
        self.clipboard_direction = opts.clipboard_direction
        self.clipboard_filter_file = opts.clipboard_filter_file

    def setup(self) -> None:
        self.init_clipboard()

    def cleanup(self) -> None:
        ch = self._clipboard_helper
        if ch:
            self._clipboard_helper = None
            ch.cleanup()

    def cleanup_protocol(self, protocol) -> None:
        ch = self._clipboard_helper
        if ch and self._clipboard_client and self._clipboard_client.protocol == protocol:
            self._clipboard_client = None
            ch.client_reset()

    def parse_hello(self, ss, _caps, send_ui: bool) -> None:
        if send_ui and self.clipboard:
            self.parse_hello_ui_clipboard(ss)

    def get_info(self, _proto) -> dict[str, Any]:
        if self._clipboard_helper is None:
            return {}
        ci = self._clipboard_helper.get_info()
        cc = self._clipboard_client
        if cc:
            ci["client"] = cc.uuid
        return {ClipboardServer.PREFIX: ci}

    def get_server_features(self, server_source=None) -> dict[str, Any]:
        clipboard = self._clipboard_helper is not None
        log("clipboard_helper=%s, clipboard_client=%s, source=%s, clipboard=%s",
            self._clipboard_helper, self._clipboard_client, server_source, clipboard)
        if not clipboard:
            return {}
        ccaps = {
            "notifications": True,
            "selections": self._clipboards,
            "want_targets": CLIPBOARD_WANT_TARGETS,
            "greedy": CLIPBOARD_GREEDY,
            "preferred-targets": CLIPBOARD_PREFERRED_TARGETS,
            "direction": self.clipboard_direction,
        }
        log("clipboard server caps=%s", ccaps)
        return {ClipboardServer.PREFIX: ccaps}

    def init_clipboard(self) -> None:
        log("init_clipboard() enabled=%s, filter file=%s", self.clipboard, self.clipboard_filter_file)
        # Clipboard handling:
        if not self.clipboard:
            return
        clipboard_filter_res = []
        if self.clipboard_filter_file:
            if not os.path.exists(self.clipboard_filter_file):
                log.error("invalid clipboard filter file: '%s' does not exist - clipboard disabled!",
                          self.clipboard_filter_file)
                return
            try:
                with open(self.clipboard_filter_file, encoding="utf8") as f:
                    for line in f:
                        clipboard_filter_res.append(line.strip())
                    log("loaded %s regular expressions from clipboard filter file %s",
                        len(clipboard_filter_res), self.clipboard_filter_file)
            except OSError:
                log.error("Error: reading clipboard filter file %s - clipboard disabled!",
                          self.clipboard_filter_file, exc_info=True)
                return
        try:
            from xpra.platform.gui import get_clipboard_native_class
            clipboard_class = get_clipboard_native_class()
            if not clipboard_class:
                raise RuntimeError("no native clipboard support on this platform")
            parts = clipboard_class.split(".")
            mod = ".".join(parts[:-1])
            class_name = parts[-1]
            module = import_module(mod)
            ClipboardClass = getattr(module, class_name)
            log("ClipboardClass for %s: %s", clipboard_class, ClipboardClass)
            kwargs = {
                "filters": clipboard_filter_res,
                "can-send": self.clipboard_direction in ("to-client", "both"),
                "can-receive": self.clipboard_direction in ("to-server", "both"),
            }
            self._clipboard_helper = ClipboardClass(self.send_clipboard_packet,
                                                    self.clipboard_progress, **kwargs)
            self._clipboard_helper.init_proxies_claim()
            self._clipboards = CLIPBOARDS
        except Exception:
            # log("gdk clipboard helper failure", exc_info=True)
            log.error("Error: failed to setup clipboard helper", exc_info=True)
            self.clipboard = False

    def parse_hello_ui_clipboard(self, ss) -> None:
        # take the clipboard if no-one else has it yet:
        if not getattr(ss, "clipboard_enabled", False):
            log("client does not support clipboard")
            return
        if not self._clipboard_helper:
            log("server does not support clipboard")
            return
        cc = self._clipboard_client
        if cc and not cc.is_closed():
            log("another client already owns the clipboard")
            return
        self.set_clipboard_source(ss)

    def set_clipboard_source(self, ss) -> None:
        if not getattr(ss, "clipboard_enabled", False):
            # don't use this client as clipboard source!
            # (its clipboard is disabled)
            return
        if self._clipboard_client == ss:
            return
        self._clipboard_client = ss
        ch = self._clipboard_helper
        log("client %s is the clipboard peer, helper=%s", ss, ch)
        if not ch:
            return
        if ss:
            log(" greedy=%s", ss.clipboard_greedy)
            log(" want targets=%s", ss.clipboard_want_targets)
            log(" server has selections: %s", csv(self._clipboards))
            log(" client initial selections: %s", csv(ss.clipboard_selections))
            ch.set_greedy_client(ss.clipboard_greedy)
            ch.set_want_targets_client(ss.clipboard_want_targets)
            ch.enable_selections(ss.clipboard_selections)
            ch.set_preferred_targets(ss.clipboard_preferred_targets)
            ch.send_tokens(ss.clipboard_selections)
        else:
            ch.enable_selections()

    def last_client_exited(self) -> None:
        ch = self._clipboard_helper
        if ch:
            ch.client_reset()

    def set_session_driver(self, source) -> None:
        self.set_clipboard_source(source)

    def _process_clipboard_packet(self, proto, packet: PacketType) -> None:
        assert self.clipboard
        if self.readonly:
            return
        ss = self.get_server_source(proto)
        if not ss:
            # protocol has been dropped!
            return
        packet_type = packet[0]
        if packet_type in ("clipboard-status", "set-clipboard-enabled"):
            self._process_clipboard_status(proto, packet)
            return
        if self._clipboard_client != ss:
            log("the clipboard packet '%s' does not come from the clipboard owner!", packet[0])
            return
        if not ss.clipboard_enabled:
            # this can happen when we disable clipboard in the middle of transfers
            # (especially when there is a clipboard loop)
            log.warn("Warning: unexpected clipboard packet")
            log.warn(" clipboard is disabled for %r", ss.uuid)
            return
        ch = self._clipboard_helper
        assert ch, "received a clipboard packet but clipboard sharing is disabled"
        GLib.idle_add(ch.process_clipboard_packet, packet)

    def _process_clipboard_status(self, proto, packet: PacketType) -> None:
        assert self.clipboard
        if self.readonly:
            return
        clipboard_enabled = packet[1]
        ss = self.get_server_source(proto)
        if ss:
            self.set_clipboard_enabled_status(ss, clipboard_enabled)

    def set_clipboard_enabled_status(self, ss, clipboard_enabled: bool) -> None:
        ch = self._clipboard_helper
        if not ch:
            log.warn("Warning: client try to toggle clipboard-enabled status,")
            log.warn(" but we do not support clipboard at all! Ignoring it.")
            return
        cc = self._clipboard_client
        if not cc:
            return
        cc.clipboard_enabled = clipboard_enabled
        log("toggled clipboard to %s for %s", clipboard_enabled, ss.protocol)
        if cc != ss or ss is None:
            log("received a request to change the clipboard status,")
            log(" but it does not come from the clipboard owner! Ignoring it.")
            log(" from %s", cc)
            log(" owner is %s", self._clipboard_client)
            return
        if not clipboard_enabled:
            ch.enable_selections()

    def clipboard_progress(self, local_requests: int, _remote_requests: int) -> None:
        assert self.clipboard
        if self._clipboard_client:
            self._clipboard_client.send_clipboard_progress(local_requests)

    def send_clipboard_packet(self, packet_type, *parts: PacketElement) -> None:
        assert self.clipboard
        if self._clipboard_client:
            packet = (packet_type, *parts)
            self._clipboard_client.send_clipboard(packet)

    def init_packet_handlers(self) -> None:
        if self.clipboard:
            for x in (
                    "token", "request", "contents", "contents-none",
                    "pending-requests", "enable-selections", "loop-uuids",
                    "status",
            ):
                self.add_packet_handler(f"{ClipboardServer.PREFIX}-%s" % x, self._process_clipboard_packet)
            self.add_legacy_alias("set-clipboard-enabled", f"{ClipboardServer.PREFIX}-status")



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.platform.win32 import constants as win32con
from xpra.server.keyboard_config_base import KeyboardConfigBase
from xpra.platform.win32.common import MapVirtualKeyW, GetAsyncKeyState, VkKeyScanW, keybd_event
from xpra.log import Logger

log = Logger("keyboard")

MAPVK_VK_TO_VSC = 0


def fake_key(keycode, press):
    if keycode <= 0:
        log.warn("no keycode found for %s", keycode)
        return
    # KEYEVENTF_SILENT = 0X4
    flags = 0
    if not press:
        flags |= win32con.KEYEVENTF_KEYUP
    # get the scancode:
    scancode = MapVirtualKeyW(keycode, MAPVK_VK_TO_VSC)
    # see: http://msdn.microsoft.com/en-us/library/windows/desktop/ms646304(v=vs.85).aspx
    log("fake_key(%s, %s) calling keybd_event(%s, %s, %s, 0)", keycode, press, keycode, scancode, flags)
    keybd_event(keycode, scancode, flags, 0)


class KeyboardConfig(KeyboardConfigBase):

    def __repr__(self):
        return "win32.KeyboardConfig"

    def do_get_keycode(self, client_keycode, keyname, pressed, modifiers, keyval, keystr, group):
        keycode = KEYCODES.get(keyname, -1)
        if keycode == -1 and keystr and len(keystr) == 1:
            v = VkKeyScanW(keystr)
            vk_code = v & 0xff
            if vk_code > 0 and vk_code != 0xff:
                keycode = vk_code
        log("get_keycode%s=%s", (client_keycode, keyname, pressed, modifiers, keyval, keystr, group), keycode)
        return keycode, group

    def make_keymask_match(self, modifier_list, ignored_modifier_keycode=None, ignored_modifier_keynames=None):
        log("make_keymask_match%s", (modifier_list, ignored_modifier_keycode, ignored_modifier_keynames))
        log("keys pressed=%s", ",".join(str(VK_NAMES.get(i, i)) for i in range(256) if GetAsyncKeyState(i) > 0))
        current = set(self.get_current_mask())
        wanted = set(modifier_list or [])
        log("make_keymask_match: current mask=%s, wanted=%s, ignoring=%s/%s",
            current, wanted, ignored_modifier_keycode, ignored_modifier_keynames)
        if current == wanted:
            return

        def is_ignored(modifier):
            if not ignored_modifier_keynames:
                return False
            for keyname in ignored_modifier_keynames:  # ie: ["Control_R"]
                keycode = KEYCODES.get(keyname, 0)  # ie: "Control_R" -> VK_RCONTROL
                if keycode > 0:
                    key_mod = MOD_KEYS.get(keycode)  # ie: "control"
                    if key_mod == modifier:
                        return True
            return False  # not found

        def change_mask(modifiers, press, info) -> None:
            for modifier in modifiers:
                if is_ignored(modifier):
                    log("change_mask: ignoring %s", modifier)
                    continue
                # find the keycode:
                for k, v in MOD_KEYS.items():
                    if ignored_modifier_keycode and ignored_modifier_keycode == k:
                        log("change_mask: ignoring %s / %s", VK_NAMES.get(k, k), v)
                        continue
                    if v == modifier:
                        # figure out if this is the one that needs toggling:
                        is_pressed = GetAsyncKeyState(k)
                        log("make_keymask_match: %s pressed=%s", k, is_pressed)
                        if bool(is_pressed) != press:
                            log("make_keymask_match: using %s to %s %s", VK_NAMES.get(k, k), info, modifier)
                            fake_key(k, press)
                            break

        change_mask(current.difference(wanted), False, "remove")
        change_mask(wanted.difference(current), True, "add")

    def get_current_mask(self):
        mods = set()
        for vk, mod in MOD_KEYS.items():
            if GetAsyncKeyState(vk) != 0:
                mods.add(mod)
        return list(mods)


MOD_KEYS = {
    win32con.VK_LSHIFT: "shift",
    win32con.VK_RSHIFT: "shift",
    win32con.VK_LCONTROL: "control",
    win32con.VK_RCONTROL: "control",
    win32con.VK_CAPITAL: "lock",
    win32con.VK_LMENU: "mod1",  # Alt_L
    win32con.VK_RMENU: "mod1",  # Alt_R
    win32con.VK_NUMLOCK: "num",
}

# we currently assume that all key events are sent using X11 names,
# so we need to translate them to win32 VK constants
# even when the client and server are both win32...
# we should support native untranslated win32 values,
# either using a special capability,
# or by adding a new argument to keypress packets
VIRTUAL_KEYS = [
    # "Mappable" codes
    ("EXECUTE", "Execute"),
    ("EXSEL", "Ex"),
    # no useful mapping:
    ("ICO_CLEAR", "IcoClr"),
    ("ICO_HELP", "Help"),
    ("DIVIDE", "KP_Divide"),
    ("MULTIPLY", "KP_Multiply"),
    ("SUBTRACT", "KP_Subtract"),
    ("ADD", "KP_Add"),
    ("NONAME", "NoSymbol"),
    ("RETURN", "KP_Enter"),
    ("NUMPAD0", "KP_0"),
    ("NUMPAD1", "KP_1"),
    ("NUMPAD2", "KP_2"),
    ("NUMPAD3", "KP_3"),
    ("NUMPAD4", "KP_4"),
    ("NUMPAD5", "KP_5"),
    ("NUMPAD6", "KP_6"),
    ("NUMPAD7", "KP_7"),
    ("NUMPAD8", "KP_8"),
    ("NUMPAD9", "KP_9"),
    ("OEM_PERIOD", "KP_Decimal"),
    ("OEM_1", "OEM_1"),
    ("OEM_102", "OEM_102"),
    ("OEM_2", "OEM_2"),
    ("OEM_3", "OEM_3"),
    ("OEM_4", "OEM_4"),
    ("OEM_5", "OEM_5"),
    ("OEM_6", "OEM_6"),
    ("OEM_7", "OEM_7"),
    ("OEM_8", "OEM_8"),
    ("OEM_ATTN", "Oem"),
    ("OEM_AUTO", "Auto"),
    ("OEM_AX", "Ax"),
    # ("OEM_BACKTAB",             "BackSpace"),
    ("BACK", "BackSpace"),
    ("ESCAPE", "Escape"),
    ("OEM_CLEAR", "OemClr"),
    ("OEM_COMMA", "OEM_COMMA"),
    ("OEM_COPY", "Copy"),
    ("OEM_CUSEL", "Cu"),
    ("OEM_ENLW", "Enlw"),
    ("OEM_FINISH", "Finish"),
    ("OEM_FJ_LOYA", "Loya"),
    ("OEM_FJ_MASSHOU", "Mashu"),
    ("OEM_FJ_ROYA", "Roya"),
    ("OEM_FJ_TOUROKU", "Touroku"),
    ("OEM_JUMP", "Jump"),
    ("OEM_MINUS", "OEM_MINUS"),
    ("OEM_PA1", "OemPa1"),
    ("OEM_PA2", "OemPa2"),
    ("OEM_PA3", "OemPa3"),
    # ("OEM_PLUS"              : "equal"),
    ("OEM_RESET", "Reset"),
    ("OEM_WSCTRL", "WsCtrl"),
    ("PA1", "Pa1"),
    # ("OEM_102"               : "backslash"),
    # missing:?
    # ("PACKET"                : "Packet"),
    ("PLAY", "Play"),
    ("PROCESSKEY", "Process"),
    ("RETURN", "Return"),
    ("SELECT", "Select"),
    ("SEPARATOR", "Separator"),
    ("SPACE", "Space"),
    ("SPACE", "space"),
    ("SUBTRACT", "minus"),
    ("TAB", "Tab"),
    ("ZOOM", "Zoom"),
    # "Non-mappable" codes
    ("BROWSER_FAVORITES", "XF86Favorites"),
    ("BROWSER_FORWARD", "XF86Forward"),
    ("BROWSER_HOME", "XF86HomePage"),
    ("BROWSER_REFRESH", "XF86Reload"),
    ("BROWSER_SEARCH", "XF86Search"),
    ("BROWSER_STOP", "XF86Suspend"),
    ("CAPITAL", "Caps_Lock"),
    ("CONVERT", "Convert"),
    ("DELETE", "Delete"),
    ("DOWN", "Down"),
    ("END", "End"),
    ("F1", "F1"),
    ("F10", "F10"),
    ("F11", "F11"),
    ("F12", "F12"),
    ("F13", "F13"),
    ("F14", "F14"),
    ("F15", "F15"),
    ("F16", "F16"),
    ("F17", "F17"),
    ("F18", "F18"),
    ("F19", "F19"),
    ("F2", "F2"),
    ("F20", "F20"),
    ("F21", "F21"),
    ("F22", "F22"),
    ("F23", "F23"),
    ("F24", "F24"),
    ("F3", "F3"),
    ("F4", "F4"),
    ("F5", "F5"),
    ("F6", "F6"),
    ("F7", "F7"),
    ("F8", "F8"),
    ("F9", "F9"),
    ("FINAL", "Final"),
    ("HELP", "Help"),
    ("HOME", "Home"),
    ("ICO_00", "Ico00"),
    ("INSERT", "Insert"),
    ("JUNJA", "Junja"),
    ("KANA", "Kana"),
    ("KANJI", "Kanji"),
    ("LAUNCH_APP1", "XF86LaunchA"),
    ("LAUNCH_APP2", "XF86LaunchB"),
    ("LAUNCH_MAIL", "XF86Mail"),
    ("LAUNCH_MEDIA_SELECT", "XF86AudioMedia"),
    ("LCONTROL", "Control_L"),
    ("LEFT", "Left"),
    ("LMENU", "Alt_L"),
    ("LSHIFT", "Shift_L"),
    ("LWIN", "Super_L"),
    ("MBUTTON", "Middle"),
    ("MEDIA_NEXT_TRACK", "Next"),
    ("MEDIA_PLAY_PAUSE", "Play"),
    ("MEDIA_PREV_TRACK", "Previous"),
    ("MEDIA_STOP", "Stop"),
    ("MODECHANGE", "Mode"),
    ("NEXT", "Page"),
    ("NONCONVERT", "Non"),
    ("NUMLOCK", "Num"),
    ("OEM_FJ_JISHO", "Jisho"),
    ("PAUSE", "Pause"),
    ("PRINT", "Print"),
    ("PRIOR", "Page"),
    ("RCONTROL", "Control_R"),
    ("RIGHT", "Right"),
    ("RMENU", "Alt_R"),
    ("RSHIFT", "Shift_R"),
    ("RWIN", "Super_R"),
    ("SCROLL", "Scroll_Lock"),
    ("SLEEP", "XF86Sleep"),
    ("SNAPSHOT", "Print"),
    ("UP", "Up"),
    ("VOLUME_DOWN", "XF86AudioLowerVolume"),
    ("VOLUME_MUTE", "XF86AudioMute"),
    ("VOLUME_UP", "XF86AudioRaiseVolume"),
    ("XBUTTON1", "X1"),
    ("XBUTTON2", "X2"),
]

KEYSYM_DEFS = {
    "bracketleft": "[",
    "bracketright": "]",
    "grave": "`",
    "braceleft": "{",
    "braceright": "}",
    "colon": ":",
    "semicolon": ";",
    "apostrophe": "'",
    "at": "@",
    "numbersign": "#",
    "comma": ",",
    "less": "<",
    "equal": "=",
    "greater": ">",
    "period": ".",
    "slash": "/",
    "question": "?",
    "bar": "|",
    "exclam": "!",
    "quotedbl": '"',
    "sterling": "£",
    "dollar": "$",
    "percent": "%",
    "asciicircum": "^",
    "ampersand": "&",
    "asterisk": "*",
    "parenleft": "(",
    "parenright": ")",
    "underscore": "_",
    "backslash": "\\",
    "asciitilde": "~",
    "notsign": "¬",
    "plus": "+",
    "eacute": "é",
    "onesuperior": "¹",
    "twosuperior": "²",
    "egrave": "è",
    "ccedilla": "ç",
    "agrave": "à",
    "dead_circumflex": "^",
    "ugrave": "ù",
    "mu": "µ",
    "section": "§",
    "currency": "¤",
    "exclamdown": "¡",
    "oneeighth": "⅛",
    "threeeighths": "⅜",
    "fiveeighths": "⅝",
    "seveneighths": "⅞",
    "trademark": "™",
    "plusminus": "±",
    "degree": "°",
    "questiondown": "¿",
    "dead_ogonek": "˛",
    "dead_macron": "¯",
    "dead_abovering": "°",
    "dead_breve": "˘",
    "dead_caron": "ˇ",
    "masculine": "º",
    "dead_abovedot": "˙",
    "division": "÷",
    "multiply": "×",
    "brokenbar": "¦",
}

# these aren't defined in win32con...
DEFS = {
    "SLEEP": 0x5F,
    "OEM_FJ_JISHO": 0x92,
    "OEM_FJ_MASSHOU": 0x93,
    "OEM_FJ_TOUROKU": 0x94,
    "OEM_FJ_LOYA": 0x95,
    "OEM_FJ_ROYA": 0x96,
    "BROWSER_SEARCH": 0xAA,
    "BROWSER_REFRESH": 0xA8,
    "BROWSER_STOP": 0xA9,
    "BROWSER_FAVORITES": 0xAB,
    "BROWSER_HOME": 0xAC,
    "MEDIA_STOP": 0xB2,
    "LAUNCH_MAIL": 0xB4,
    "LAUNCH_MEDIA_SELECT": 0xB5,
    "LAUNCH_APP1": 0xB6,
    "LAUNCH_APP2": 0xB7,
    "OEM_PLUS": 0xBB,
    "OEM_COMMA": 0xBC,
    "OEM_MINUS": 0xBD,
    "OEM_PERIOD": 0xBE,
    "OEM_1": 0xBA,
    "OEM_2": 0xBF,
    "OEM_3": 0xC0,
    "OEM_4": 0xDB,
    "OEM_5": 0xDC,
    "OEM_6": 0xDD,
    "OEM_7": 0xDE,
    "OEM_8": 0xDF,
    "OEM_AUTO": 0xF3,
    "OEM_COPY": 0xF2,
    "OEM_AX": 0xE1,
    "OEM_102": 0xE2,
    "ICO_HELP": 0xE3,
    "ICO_00": 0xE4,
    "ICO_CLEAR": 0xE6,
    "OEM_RESET": 0xE9,
    "OEM_JUMP": 0xEA,
    "OEM_PA1": 0xEB,
    "OEM_PA2": 0xEC,
    "OEM_PA3": 0xED,
    "OEM_WSCTRL": 0xEE,
    "OEM_CUSEL": 0xEF,
    "OEM_ATTN": 0xF0,
    "OEM_FINISH": 0xF1,
    "OEM_ENLW": 0xF4,
    "OEM_BACKTAB": 0xF5,
}


def init_vk_names():
    names = {}
    for name in (x for x in dir(win32con) if x.startswith("VK_")):
        names[getattr(win32con, name)] = name
    for name, val in DEFS.items():
        names[val] = "VK_" + name
    return names


VK_NAMES = init_vk_names()
log("VK_NAMES=%s", VK_NAMES)


def init_keycodes():
    # lookup the constants:
    keycodes = {}
    for vk, name in VIRTUAL_KEYS:
        vk_name = "VK_%s" % vk
        if hasattr(win32con, vk_name):
            val = getattr(win32con, vk_name)
            keycodes[name] = val
            log("KEYCODES[%s]=win32con.%s=%s", name, vk_name, val)
        elif vk in DEFS:
            # fallback to our hardcoded definitions:
            val = DEFS[vk]
            keycodes[name] = val
            log("KEYCODES[%s]=%s=%s", name, vk, val)
        else:
            log.warn("missing key constant: %s", vk_name)

    for name, char in KEYSYM_DEFS.items():
        if len(char) != 1:
            log.warn("invalid character '%s' : '%s' (len=%i)", name, char, len(char))
            continue
        v = VkKeyScanW(char)
        vk_code = v & 0xff
        if vk_code > 0 and vk_code != 0xff:
            log("KEYCODE[%s]=%i (%s)", char, vk_code, name)
            keycodes[name] = vk_code
        else:
            log("no keycode for %r (%s)", char, name)

    keycodes.update({
        "Shift_L": win32con.VK_LSHIFT,
        "Shift_R": win32con.VK_RSHIFT,
        "Control_L": win32con.VK_LCONTROL,
        "Control_R": win32con.VK_RCONTROL,
        "Caps_Lock": win32con.VK_CAPITAL,
        "Num_Lock": win32con.VK_NUMLOCK,
        "Scroll_Lock": win32con.VK_SCROLL,
    })

    for c in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
        keycodes[c] = ord(c)
        keycodes[c.lower()] = ord(c)
    for c in "0123456789":
        keycodes[c] = ord(c)
    return keycodes


# lookup the constants:
KEYCODES = init_keycodes()
log("KEYCODES: %s", KEYCODES)
